/*	FUZZauth - http://unfix.org/projects/fuzzauth & http://fuzzauth.sourceforge.net
 * FUZZauth.h - common definitions by Jeroen Massar <jeroen@unfix.org>
*/

#ifndef FUZZAUTH_H
#define FUZZAUTH_H "FUZZauth"

#define FUZZAUTH_NAME					"FUZZauth"							// Name of this service.
#define FUZZAUTH_NAME_LOW				"fuzzauth"							// Name of this service (lowercase only, for PAM/NSS modules etc.).
#define FUZZAUTH_MAGIC					"FUZZ"								// Magic identifier.
#define FUZZAUTH_AUTHOR					"Jeroen Massar"					// Who made it.
#define FUZZAUTH_EMAIL					"jeroen@unfix.org"				// The email address of the author.
#define FUZZAUTH_DATE					"2002.02.12"						// Version of the service.
#define FUZZAUTH_VERSION				1										// Protocol v1.
#define FUZZAUTH_PORT					"5042"								// Our default port.
#define FUZZAUTH_HOMEDIR				"/var/lib/fuzzauth"				// Homedir of FUZZauth (contains: bin,lib,etc).
#define FUZZAUTH_VARDIR					"/var/run/fuzzauth"				// Variable directory for socket & pid files.
#define FUZZAUTH_UNIXSOCKET			"/var/run/fuzzauth/socket"		// The UNIX socket for local clients.
#define FUZZAUTH_FUZZD_PID				"/var/run/fuzzauth/fuzzd.pid"	// The UNIX socket for local clients.

#define FUZZAUTH_WEBSITE "http://unfix.org/projects/fuzzauth and http://fuzzauth.sourceforge.net"	// Where to find it on the web.

#define FA_CRYPT_MHASH "MHASH supported"

// Encryption modes for password storage.
enum c_nEncryptionModes{
												c_nEncryption_Clear = 0,	// Cleartext (no encryption)
												c_nEncryption_Crypt,			// crypt() encryption.
												c_nEncryption_MD5,			// MHASH - MD5.
												c_nEncryption_CRC32,			// MHASH - CRC32.
												c_nEncryption_CRC32B,		// MHASH - CRC32B.
												c_nEncryption_SHA1,			// MHASH - SHA1.
												c_nEncryption_SHA256,		// MHASH - SHA256.
												c_nEncryption_HAVAL256,		// MHASH - HAVAL256.
												c_nEncryption_HAVAL192,		// MHASH - HAVAL192.
												c_nEncryption_HAVAL160,		// MHASH - HAVAL160.
												c_nEncryption_HAVAL128,		// MHASH - HAVAL128.
												c_nEncryption_RIPEMD160,	// MHASH - RIPEMD160.
												c_nEncryption_TIGER,			// MHASH - TIGER.
												c_nEncryption_TIGER160,		// MHASH - TIGER160.
												c_nEncryption_TIGER128,		// MHASH - TIGER128.
												c_nEncryption_GOST			// MHASH - GOST.
};

enum c_nMode {
												c_nMode_Unknown = 0,			// Unknown Mode, should never happen.
												c_nMode_Human,					// Human readable, no encryption (default).
												c_nMode_Packet,				// Packet mode, no encryption (default on the local UNIX socket).
												c_nMode_Packet_SSH,			// Packet mode, SSH encryption.
												c_nMode_Packet_SSL			// Packet mode, SSL encryption.
	};
#define c_sMode_Unknown					"Unknown"
#define c_sMode_Human					"Human"
#define c_sMode_Packet					"Packet"
#define c_sMode_Packet_SSH				"Packet/SSL"
#define c_sMode_Packet_SSL				"Packet/SSH"

enum c_nCommand {
												c_nCommand_Unknown = 0,		// Unknown Command, should never happen.
												c_nCommand_Quit,				// Close the connection.
												c_nCommand_Mode,				// Change Mode.

												// PAM
												c_nCommand_Authenticate,	// Try Authentication.
												c_nCommand_ChangePassword,	// Update Password.

												// NSS
												c_nCommand_GetUsers,			// Get users [getpwent()].
												c_nCommand_MatchUserID,		// Match a user by id [getpwuid()].
												c_nCommand_MatchUserName,	// Match a user by name [getpwnam()].
												c_nCommand_MatchUserGECOS,	// Match a user by GECOS.
												c_nCommand_GetGroups,		// Get groups [getgrent()].
												c_nCommand_MatchGroupID,	// Match a user by id [getgrgid()].
												c_nCommand_MatchGroupName,	// Match a user by id [getgrnam()].

												// Management
												c_nCommand_UserAdd,			// Add a new user.
												c_nCommand_UserRemove,		// Remove a user.
												c_nCommand_UserChange,		// Change a user.
												c_nCommand_GroupAdd,			// Add a new group.
												c_nCommand_GroupRemove,		// Remove a group.
												c_nCommand_GroupChange,		// Change a group.
												c_nCommand_MemberAdd,		// Add a new member to a group.
												c_nCommand_MemberRemove		// Remove a member from a group.
	};
#define c_sCommand_Unknown				"UNKNOWN"
#define c_sCommand_Quit					"QUIT"
#define c_sCommand_Mode					"MODE"
#define c_sCommand_Authenticate		"AUTHENTICATE"
#define c_sCommand_ChangePassword	"CHANGEPASSWORD"

#define c_sCommand_GetUsers			"GETUSERS"
#define c_sCommand_MatchUserID		"MATCHUSERID"
#define c_sCommand_MatchUserName		"MATCHUSERNAME"
#define c_sCommand_MatchUserGECOS	"MATCHUSERGECOS"
#define c_sCommand_GetGroups			"GETGROUPS"
#define c_sCommand_MatchGroupID		"MATCHGROUPID"
#define c_sCommand_MatchGroupName	"MATCHGROUPNAME"

#define c_sCommand_UserAdd				"USERADD"
#define c_sCommand_UserRemove			"USERREMOVE"
#define c_sCommand_UserChange			"USERCHANGE"
#define c_sCommand_GroupAdd			"GROUPADD"
#define c_sCommand_GroupRemove		"GROUPREMOVE"
#define c_sCommand_GroupChange		"GROUPCHANGE"
#define c_sCommand_MemberAdd			"MEMBERADD"
#define c_sCommand_MemberRemove		"MEMBERREMOVE"

enum c_nAnswer {
												c_nAnswer_Unknown = 0,		// Unknown Answer, should never happen.
												c_nAnswer_UnknownVersion,	// Unknown Protocol version.
												c_nAnswer_UnknownMode,		// Unknown Mode.
												c_nAnswer_UnknownCommand,	// Unknown Command.
												c_nAnswer_NeedDataPacket,	// Need Data Packet (used internaly).
												c_nAnswer_AlreadySent,		// Answer already sent.
												c_nAnswer_NoMoreData,		// No more data packets (previous was the last one).
												c_nAnswer_Granted,			// Request Granted.
												c_nAnswer_Denied				// Request Denied.
	};
#define c_sAnswer_Unknown				"Unknown"
#define c_sAnswer_UnknownVersion		"Unknown Version"
#define c_sAnswer_UnknownMode			"Unknown Mode"
#define c_sAnswer_UnknownCommand		"Unknown Command"
#define c_sAnswer_NeedDataPacket		"Need Data Packet"
#define c_sAnswer_AlreadySent			"Already Sent"
#define c_sAnswer_NoMoreData			"No More Data"
#define c_sAnswer_Granted				"Granted"
#define c_sAnswer_Denied				"Denied"

enum c_nDatatype {
												c_nDatatype_Unknown = 0,	// Unknown Datatype (=c_nDatatype_None)
												c_nDatatype_Number,			// faPacket_Number
												c_nDatatype_Number2,			// faPacket_Number2
												c_nDatatype_String,			// faPacket_String
												c_nDatatype_User,				// faPacket_User
												c_nDatatype_Group				// faPacket_Group
	};
#define c_nDatatype_None				c_nDatatype_Unknown
#define c_sDatatype_Unknown			"Unknown"
#define c_sDatatype_Number				"Number"
#define c_sDatatype_Number2			"Number2"
#define c_sDatatype_String				"String"
#define c_sDatatype_User				"User"
#define c_sDatatype_Group				"Group"

struct faPacket_Request
{
	char										m_nMagic[4];					// Magic identifier.
	unsigned int							m_nVersion;						// Protocol version number.
	enum c_nCommand						m_nCommand;						// Command.
	char 										m_sAdminName[100];			// Administrator Name (*).
	char										m_sAdminPassword[100];		// Administrator Password (*).
	enum c_nDatatype						m_nDatatype;					// Type of the data following this.
	unsigned int							m_nLength;						// Length of the data packet which is following this.
};

// (*) = These two Admin fields are required for some operations.

struct faPacket_Answer
{
	enum c_nAnswer							m_nAnswer;						// Answer.
	enum c_nDatatype						m_nDatatype;					// Answer data type.
	unsigned int							m_nLength;						// Length of data packet following this one (0 -> no data packet).
};

struct faPacket_Number
{
	unsigned int							m_nNumber;						// Number.
};

// Used for Member Add/Remove
struct faPacket_Number2
{
	unsigned int							m_nNumber1;						// Number 1.
	unsigned int							m_nNumber2;						// Number 2.
};

struct faPacket_String
{
	char										m_sString[100];				// String.
};

struct faPacket_User
{
	char										m_sUsername[100];				// User name.
	char										m_sPassword[100];				// Password.
	char										m_sPasswordNew[100];			// New Password.
	unsigned int							m_nUID;							// User ID.
	unsigned int							m_nGID;							// Primary Group ID.
	char										m_sGECOS[100];					// GECOS.
	char										m_sHomedir[100];				// Homedir.
	char										m_sShell[100];					// Shell - without path, which is taken from /etc/shells.
};

struct faPacket_Group
{
	unsigned int							m_nGID;							// Group ID.
	char										m_sGroupname[100];			// Group name.
	char										m_sPassword[100];				// Password.
	unsigned int							m_nNumMembers;					// Number of members in this group.
};

#ifndef bool
#define bool int
#endif
#ifndef false
#define false 0
#endif
#ifndef true
#define true (!false)
#endif

#endif // ifndef FUZZAUTH_H
