/*	FUZZauth - http://unfix.org/projects/fuzzauth & http://fuzzauth.sourceforge.net
 * FUZZauth_common.h - common interface definitions by Jeroen Massar <jeroen@unfix.org>
*/
#ifndef FUZZAUTH_COMMON_H
#define FUZZAUTH_COMMON_H "FUZZauth"

#ifdef WIN32
#define ERRORNUMSTR WSAGetLastError(),DecodeError(WSAGetLastError())
#else
#define SOCKET int
#define INVALID_SOCKET (-1)
#define SOCKET_ERROR (-1)
#define closesocket close
#define ERRORNUMSTR errno,strerror(errno)
#endif

#define __STRICT_ANSI__ 1
#define _GNU_SOURCE 1
#define _LARGEFILE64_SOURCE 1
#define _REENTRANT 1

#include <features.h>

#ifndef __GNUC__
#define __attribute__(x) /*nothing*/
#endif

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <stdarg.h>
#include <sys/types.h>
#include <fcntl.h>

#ifndef WIN32
#include <unistd.h>
#include <netdb.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <signal.h>
#endif
#include <sys/un.h>
#include <pwd.h>
#include <grp.h>
#include <sys/stat.h>

// FUZZauth
#include "fuzzauth.h"

#ifndef D
#ifdef DEBUG
#define D(x) x;
void DLOG(const char *sFormat, ... );
#else
#define D(x) ;
#endif
#endif

#define FA_PROMPT_USERNAME				"New FUZZauth password: "
#define FA_PROMPT_PASSWORD_NEW		"New FUZZauth password: "
#define FA_PROMPT_PASSWORD_CONFIRM	"Confirm new FUZZauth password: "
#define FA_PROMPT_PASSWORD_MISMATCH	"Sorry, passwords do not match"
#define FA_PASSWORD_CHANGING			"Changing FUZZauth password for %s"

// Public Function Prototypes
void	fuzzauth_LogA(int nPriority, const char *sFormat, va_list xArgs) __attribute__((format(printf,2,0)));
void	fuzzauth_Log(int nPriority, const char *sFormat, ... ) __attribute__((format(printf,2,3)));
bool	fuzzauth_ConnectA(SOCKET *sSocket, int *nConnectCount);
void	fuzzauth_DisconnectA(SOCKET *sSocket, int *nConnectCount);
bool	fuzzauth_Connect(void);
bool	fuzzauth_ConnectB(SOCKET *sSocket, int *nConnectCount, char *sHost, char *sPort, int nFamily);
void	fuzzauth_Disconnect(void);

// Commands
// MODE is not implemented here as UNIX are local and really don't need any encryption.
bool	fuzzauth_Authenticate(const char *sUsername, const char *sPassword);
bool	fuzzauth_ChangePassword(char *sUsername, char *sPasswordOld, char *sPasswordNew);
bool	fuzzauth_UserExists(const char *sUsername);

// Getting Answer Packets.
//bool fuzzauth_GetAnswer(int nDatatype);
bool	fuzzauth_GetUsers(bool bFirst, struct passwd *xPasswd, char *sBuffer, size_t nBuflen);
bool	fuzzauth_MatchUserName(const char *sname, struct passwd *xPasswd, char *sBuffer, size_t nBuflen);
bool	fuzzauth_MatchUserID(int nUID, struct passwd *xPasswd, char *sBuffer, size_t nBuflen);
int	fuzzauth_GetGroups(bool bFirst, struct group *xGroup, char *sBuffer, size_t nBuflen);
bool	fuzzauth_MatchGroupName(const char *sname, struct group *xGroup, char *sBuffer, size_t nBuflen);
bool	fuzzauth_MatchGroupID(int nGID, struct group *xGroup, char *sBuffer, size_t nBuflen);

// Management
bool	fuzzauth_UserAdd(char *sUsername, char *sPassword, char *sPrimaryGroup, char *sGECOS, char *sHomedir, char *sShell);
bool	fuzzauth_UserRemove(char *sUsername);
bool	fuzzauth_UserChange(char *sUsername, char *sNewUsername, char *sPassword);
bool	fuzzauth_GroupAdd(char *sGroupname, char *sPassword);
bool	fuzzauth_GroupRemove(char *sGroupname);
bool	fuzzauth_GroupChange(char *sGroupname, char *sNewGroupname, char *sPassword);
bool	fuzzauth_MemberAdd(char *sUsername, char *sGroupname);
bool	fuzzauth_MemberRemove(char *sUsername, char *sGroupname);

#endif // ifndef FUZZAUTH_COMMON_H
