/* FUZZd MySQL Authenticator by Jeroen Massar <jeroen@unfix.org> */

// FUZZd
#include "fuzzd.h"
#include "fuzzauth_config.h"

// MySQL Includes
#include <mysql/mysql.h>

// Globals
MYSQL g_xConn;

struct Config g_xConfig[] ={
										{"Database",								NULL, 0},
										{"Servername",								NULL, 0},
										{"Username",								NULL, 0},
										{"Password",								NULL, 0},
										// Queries
										{"Query_Password_From_User",			NULL, 0},
										{"Query_Update_Password",				NULL, 0},
										{"Query_All_Users",						NULL, 0},
										{"Query_User_From_UserID",				NULL, 0},
										{"Query_User_From_Username",			NULL, 0},
										{"Query_User_From_GECOS",				NULL, 0},
										{"Query_GroupMember_From_GroupID",	NULL, 0},
										{"Query_All_Groups",						NULL, 0},
										{"Query_Group_From_GroupID",			NULL, 0},
										{"Query_Group_From_Groupname",		NULL, 0},
										// UserAdd()
										{"UserAdd_Script",						NULL, 0},
										{"Query_UserAdd_User",					NULL, 0},
										{"Query_UserAdd_Group",					NULL, 0},
										{"Query_UserAdd_Password",				NULL, 0},
										// UserRemove()
										{"UserRemove_Script",					NULL, 0},
										{"Query_UserRemove_User",				NULL, 0},
										{"Query_UserRemove_Group",				NULL, 0},
										{"Query_UserRemove_Password",			NULL, 0}
									};

enum c_nConfig {
										c_nDatabase = 0,
										c_nServername,
										c_nUsername,
										c_nPassword,
										// Queries
										c_nQuery_Password_From_User,
										c_nQuery_Update_Password,
										c_nQuery_All_Users,
										c_nQuery_User_From_UserID,
										c_nQuery_User_From_Username,
										c_nQuery_User_From_GECOS,
										c_nQuery_GroupMember_From_GroupID,
										c_nQuery_All_Groups,
										c_nQuery_Group_From_GroupID,
										c_nQuery_Group_From_Groupname,
										// UserAdd()
										c_nUserAdd_Script,
										c_nQuery_UserAdd_User,
										c_nQuery_UserAdd_UserID,
										c_nQuery_UserAdd_Group,
										c_nQuery_UserAdd_Password,
										c_nUserRemove_Script,
										c_nQuery_UserRemove_User,
										c_nQuery_UserRemove_Group,
										c_nQuery_UserRemove_Password
					};

/***********************
 Initialize the module
***********************/
bool MY_Initialize(void)
{
	bool				bRet = false;
	unsigned int	i;
	char				*sError = NULL;

	D(DLOG("MY_Initialize()\n"));
	if ((sError = ParseConfig("auth_mysql", g_xConfig, sizeof(g_xConfig)/sizeof(struct Config))))
	{
		for (bRet = true, i = 0; bRet && (i<sizeof(g_xConfig)/sizeof(struct Config)); i++)
		{
//			D(DLOG("Checking %d: %s -> %s\n", i, g_xConfig[i].m_sName, g_xConfig[i].m_sValue));
			if (g_xConfig[i].m_sValue == NULL) bRet = false;
		}
		if (bRet)
		{
			bRet = false;
			mysql_init(&g_xConn);

			if (mysql_real_connect(&g_xConn, g_xConfig[c_nServername].m_sValue, g_xConfig[c_nUsername].m_sValue, g_xConfig[c_nPassword].m_sValue, g_xConfig[c_nDatabase].m_sValue, 0, NULL, 0))
			{
				fuzzd_Log(LOG_INFO, "MySQL connection succesful.\n");
			}
			else
			{
				fuzzd_Log(LOG_ERR, "MySQL connection failed: '%s'\n", mysql_error(&g_xConn));
				bRet = false;
			}
		}
		else
		{
			i--;
			fuzzd_Log(LOG_ERR, "Unset option ('%s') detected ...\n", g_xConfig[i].m_sName);
		}
	}
	else
	{
		fuzzd_Log(LOG_ERR, "%s", sError);
		free(sError);
	}
	return bRet;
}

/*********************
 Shutdown the module
*********************/
bool MY_Shutdown(void)
{
	D(DLOG("MY_Shutdown()\n"));
	mysql_close(&g_xConn);
	return true;
}

/******************************
 Get some information.
******************************/
bool MY_GetInfo(char *sBuf, int nSize)
{
	strncpy(sBuf, "MySQL FUZZauth Authenticator Plugin by Jeroen Massar **** UNTESTED ****\n", nSize);
	return true;
}

/******************************
 (local) Execute an SQL Query.
******************************/
bool MY_Exec(struct Client *xClient, MYSQL_RES **xResult, char *sQuery)
{
	bool bRet = false;

	D(DLOG("[%s] MY_Exec(%s)\n", xClient->m_sHostname, sQuery));
	if (!mysql_query(&g_xConn, sQuery))
	{
		*xResult = mysql_use_result(&g_xConn);
		if (*xResult)
		{
			//D(DLOG("[%s] MySQL query succesful\n", xClient->m_sHostname));
			bRet = true;
		}
		else
		{
			fuzzd_Log(LOG_ERR, "[%s] Couldn't get MySQL resultset: '%s'\n", xClient->m_sHostname, mysql_error(&g_xConn));
		}
	}
	else
	{
		fuzzd_Log(LOG_ERR, "[%s] MySQL query failed: '%s'\n", xClient->m_sHostname, mysql_error(&g_xConn));
	}
	return bRet;
}

/******************************************
 (local) Execute an SQL Query (Formatted).
******************************************/
bool MY_ExecA(struct Client *xClient, MYSQL_RES **xResult, char *sFormat, ...)
{
	bool		bRet = false;
	int		nLen;
	char		*sQuery = NULL;
	va_list	xArgs;

	va_start(xArgs, sFormat);
	nLen = vasprintf(&sQuery, sFormat, xArgs);
	if (sQuery)
	{
		bRet = MY_Exec(xClient, xResult, sQuery);
		free(sQuery);
	}
	va_end(xArgs);
	return bRet;
}

/**************************
 Escape all SQL-escapes.
 the result must be free()'d by the caller.
**************************/
char *SQLEscape(char *sString)
{
	int	i = 0, j = 0, nEscapes = 0, nNewLen = 0;
	char	*sEscaped = NULL;

	while (sString[i] != '\0')
	{
		if (	(sString[i] == '\'') ||
				(sString[i] == '\"') ||
				(sString[i] == '\\'))
		{
			nEscapes++;
		}
		i++;
	}

	nNewLen = i + nEscapes;
	D(DLOG("SQLEscape() - Old = %d/%d, New = %d\n", strlen(sString), i, nNewLen));

	if ((sEscaped = malloc(nNewLen + 1)))
	{
		i = 0;
		while (sString[i] != '\0')
		{
			if (	(sString[i] == '\'') ||
					(sString[i] == '\"') ||
					(sString[i] == '\\'))
			{
				sEscaped[j] = '\\';
				j++;
				sEscaped[j] = sString[i];
			}
			else sEscaped[j] = sString[i];
			j++;
			i++;
		}
		// Extra verification.
		if (j == (nNewLen))
		{
			sEscaped[j] = '\0';
			D(DLOG("SQLEscape() - \"%s\" (%d) -> \"%s\" (%d/%d)\n", sString, strlen(sString), sEscaped, strlen(sEscaped), nNewLen));
		}
		else
		{
			D(DLOG("SQLEscape() - Process failed (j=%d, nNewLen=%d).\n", j, nNewLen));
			free(sEscaped);
			sEscaped = NULL;
		}
	}
	return sEscaped;
}

/**************************
 Authenticate this Client
**************************/
bool MY_Authenticate(struct Client *xClient)
{
	bool			bRet = false;
	MYSQL_RES		*xResult = NULL;
	char			*sUsername = NULL;

	D(DLOG("[%s] MY_Authenticate()\n", xClient->m_sHostname));
	if ((sUsername = SQLEscape(xClient->m_xPkt_User.m_sUsername)))
	{
		if (MY_ExecA(xClient, &xResult, g_xConfig[c_nQuery_Password_From_User].m_sValue, sUsername))
		{
			if (mysql_num_rows(xResult) == 1)
			{
				MYSQL_ROW xRow = mysql_fetch_row(xResult);
				char *sPassword = xRow[0];
				if (sPassword)
				{
					// If we are being called by ChangePassword, also allow setting a password if it's empty...
					// This allows an administrator to clear the password in the database and then quickly update it.
					D(DLOG("[%s] MY_Authenticate() - Password Length = %d, Cmd=%d\n", xClient->m_sHostname, strlen(sPassword), xClient->m_xPkt_Request.m_nCommand));
					if ((xClient->m_xPkt_Request.m_nCommand == c_nCommand_ChangePassword) && (sPassword) && (strlen(sPassword) == 0))
					{
						D(DLOG("[%s] MY_Authenticate() - Succeeded because of ChangePassword and the database password is empty.\n", xClient->m_sHostname));
						bRet = true;
					}
					// And in any case try to compare the passwords like usual.
					else if ((sPassword) && (fuzzd_ComparePassword(sPassword, xClient->m_xPkt_User.m_sPassword)))
					{
						D(DLOG("[%s] MY_Authenticate() - Passwords match.\n", xClient->m_sHostname));
						bRet = true;
					}
					D(else DLOG("[%s] MY_Authenticate() - Passwords MISmatch.\n", xClient->m_sHostname));
				}
				D(else DLOG("[%s] MY_Authenticate() - No password in database.\n", xClient->m_sHostname));
			}
			else
			{
				D(DLOG("[%s] MY_Authenticate() - No tuples returned.\n", xClient->m_sHostname));
				bRet = false;
			}
			mysql_free_result(xResult);
		}
		free(sUsername);
	}
	else fuzzd_Log(LOG_ERR, "[%s] Couldn't allocate memory for escaping username\n", xClient->m_sHostname);
	if (xResult) mysql_free_result(xResult);
	return bRet;
}

/****************
 Change Password
****************/
bool MY_ChangePassword(struct Client *xClient)
{
	bool bRet = false;
	char *sCrypt;

	D(DLOG("[%s] MY_ChangePassword()\n", xClient->m_sHostname));
	if ((sCrypt = fuzzd_Encrypt(xClient->m_xPkt_User.m_sPasswordNew)))
	{
		MYSQL_RES *xResult = NULL;
		char		*sUsername = SQLEscape(xClient->m_xPkt_User.m_sUsername);
		if (sUsername)
		{
			char *sCryptSQL = SQLEscape(sCrypt);
			if (sCryptSQL)
			{
				if (MY_ExecA(xClient, &xResult, g_xConfig[c_nQuery_Update_Password].m_sValue, sCryptSQL, sUsername))
				{
					fuzzd_Log(LOG_INFO, "[%s] Password for user %s changed succesfully.\n", xClient->m_sHostname, xClient->m_xPkt_User.m_sUsername);
					bRet = true;
				}
				else fuzzd_Log(LOG_ERR, "[%s] Password changing for user %s failed: %s.\n", xClient->m_sHostname, xClient->m_xPkt_User.m_sUsername, mysql_error(&g_xConn));
				free(sCryptSQL);
			}
			else fuzzd_Log(LOG_ERR, "[%s] Couldn't allocate memory for escaping password\n", xClient->m_sHostname);
			free(sUsername);
		}
		else fuzzd_Log(LOG_ERR, "[%s] Couldn't allocate memory for escaping username\n", xClient->m_sHostname);
		mysql_free_result(xResult);
	}
	return bRet;
}

/********************************
 Return a User from SQL results.
********************************/
bool MY_ReturnUser(struct Client *xClient, MYSQL_RES *xResult)
{
	bool			bRet = false;
	char			*sBuf;
	MYSQL_ROW	xRow;

	D(DLOG("[%s] MY_ReturnUser()\n", xClient->m_sHostname));
	// Clear the complete thing.
	memset(&xClient->m_xPkt_User, 0, sizeof(xClient->m_xPkt_User));
	xRow = mysql_fetch_row(xResult);
	if ((sBuf = xRow[0]))
	{
		strncpy(xClient->m_xPkt_User.m_sUsername, sBuf, sizeof(xClient->m_xPkt_User.m_sUsername)-1);
		if ((sBuf = xRow[1]))
		{
			// We don't return passwords.
			//strncpy(xClient->m_xPkt_User.m_sPassword, sBuf, sizeof(xClient->m_xPkt_User.m_sPassword)-1);
			strncpy(xClient->m_xPkt_User.m_sPassword, "*", sizeof(xClient->m_xPkt_User.m_sPassword)-1);
			xClient->m_xPkt_User.m_nUID = atoi(xRow[2]);
			xClient->m_xPkt_User.m_nGID = atoi(xRow[3]);
			if ((sBuf = xRow[4]))
			{
				strncpy(xClient->m_xPkt_User.m_sGECOS,	sBuf, sizeof(xClient->m_xPkt_User.m_sGECOS)-1);
				if ((sBuf = xRow[5]))
				{
					strncpy(xClient->m_xPkt_User.m_sHomedir, sBuf, sizeof(xClient->m_xPkt_User.m_sHomedir)-1);
					if ((sBuf = xRow[6]))
					{
						D(DLOG("[%s] MY_ReturnUser() - %d:%s:%s\n", xClient->m_sHostname, xClient->m_xPkt_User.m_nUID, xClient->m_xPkt_User.m_sUsername, xClient->m_xPkt_User.m_sGECOS));
						strncpy(xClient->m_xPkt_User.m_sShell,	sBuf, sizeof(xClient->m_xPkt_User.m_sShell)-1);
						xClient->m_xPkt_Answer.m_nAnswer = c_nAnswer_Granted;
						xClient->m_xPkt_Answer.m_nDatatype = c_nDatatype_User;
						bRet = fuzzd_ReturnAnswer(xClient);
					}
					D(else DLOG("[%s] MY_GetUsers() - No value #7\n", xClient->m_sHostname));
				}
				D(else DLOG("[%s] MY_GetUsers() - No value #6\n", xClient->m_sHostname));
			}
			D(else DLOG("[%s] MY_GetUsers() - No value #5\n", xClient->m_sHostname));
		}
		D(else DLOG("[%s] MY_GetUsers() - No value #2\n", xClient->m_sHostname));
	}
	D(else DLOG("[%s] MY_GetUsers() - No value #1\n", xClient->m_sHostname));
	return bRet;
}

/****************************
 Get Users from our backend.
****************************/
bool MY_GetUsers(struct Client *xClient)
{
	bool		bRet = false;
	MYSQL_RES	*xResult = NULL;

	D(DLOG("[%s] MY_GetUsers()\n", xClient->m_sHostname));
	if (MY_ExecA(xClient, &xResult, g_xConfig[c_nQuery_All_Users].m_sValue))
	{
		int	i, nMax = mysql_num_rows(xResult);

		D(DLOG("[%s] MY_GetUsers() - Returning %d users\n", xClient->m_sHostname, nMax));
		bRet = true;
		for (i = 0; (bRet == true) && (i < nMax); i++) bRet = MY_ReturnUser(xClient, xResult);
	}
	if (xResult) mysql_free_result(xResult);
	// This one will be send by the mainloop.
	xClient->m_xPkt_Answer.m_nAnswer = c_nAnswer_NoMoreData;
	return bRet;
}

/*********************
 Match User by UserID
*********************/
bool MY_MatchUserID(struct Client *xClient)
{
	bool		bRet = false;
	MYSQL_RES	*xResult = NULL;

	D(DLOG("[%s] MY_MatchUserID(%d)\n", xClient->m_sHostname, xClient->m_xPkt_Number.m_nNumber));
	if (MY_ExecA(xClient, &xResult, g_xConfig[c_nQuery_User_From_UserID].m_sValue, xClient->m_xPkt_Number.m_nNumber))
	{
		if (mysql_num_rows(xResult) == 1)
		{
			bRet = MY_ReturnUser(xClient, xResult);
			// Already Sent...
			xClient->m_xPkt_Answer.m_nAnswer = c_nAnswer_AlreadySent;
		}
	}
	if (xResult) mysql_free_result(xResult);
	if (!bRet) xClient->m_xPkt_Answer.m_nAnswer = c_nAnswer_Denied;
	return bRet;
}

/*******************
 Match User by name
*******************/
bool MY_MatchUserName(struct Client *xClient)
{
	bool		bRet = false;
	MYSQL_RES	*xResult = NULL;
	char		*sUsername = NULL;

	D(DLOG("[%s] MY_MatchUserName('%s')\n", xClient->m_sHostname, xClient->m_xPkt_String.m_sString));
	if ((sUsername = SQLEscape(xClient->m_xPkt_String.m_sString)))
	{
		if (MY_ExecA(xClient, &xResult, g_xConfig[c_nQuery_User_From_Username].m_sValue, sUsername))
		{
			if (mysql_num_rows(xResult) == 1)
			{
				bRet = MY_ReturnUser(xClient, xResult);
				// Already Sent...
				xClient->m_xPkt_Answer.m_nAnswer = c_nAnswer_AlreadySent;
			}
		}
		if (xResult) mysql_free_result(xResult);
		free(sUsername);
	}
	else fuzzd_Log(LOG_ERR, "[%s] Couldn't allocate memory for escaping username\n", xClient->m_sHostname);
	if (!bRet) xClient->m_xPkt_Answer.m_nAnswer = c_nAnswer_Denied;
	return bRet;
}

/********************
 Match User by GECOS
********************/
bool MY_MatchUserGECOS(struct Client *xClient)
{
	bool		bRet = false;
	MYSQL_RES	*xResult = NULL;
	char		*sGECOS = NULL;

	D(DLOG("[%s] MY_MatchUserGECOS('%s')\n", xClient->m_sHostname, xClient->m_xPkt_String.m_sString));
	if ((sGECOS = SQLEscape(xClient->m_xPkt_String.m_sString)))
	{
		if (MY_ExecA(xClient, &xResult, g_xConfig[c_nQuery_User_From_GECOS].m_sValue, sGECOS))
		{
			if (mysql_num_rows(xResult) == 1)
			{
				bRet = MY_ReturnUser(xClient, xResult);
				// Already Sent...
				xClient->m_xPkt_Answer.m_nAnswer = c_nAnswer_AlreadySent;
			}
		}
		if (xResult) mysql_free_result(xResult);
		free(sGECOS);
	}
	else fuzzd_Log(LOG_ERR, "[%s] Couldn't allocate memory for escaping username\n", xClient->m_sHostname);
	if (!bRet) xClient->m_xPkt_Answer.m_nAnswer = c_nAnswer_Denied;
	return bRet;
}

/********************************
 Return a Group from SQL results.
********************************/
bool MY_ReturnGroup(struct Client *xClient, MYSQL_RES *xResult)
{
	bool			bRet = false;
	char			*sBuf;
	MYSQL_RES	*xResult2 = NULL;
	MYSQL_ROW	xRow, xRow2;

	D(DLOG("[%s] MY_ReturnGroup()\n", xClient->m_sHostname));
	// Clear the complete thing.
	memset(&xClient->m_xPkt_Group, 0, sizeof(xClient->m_xPkt_Group));
	xRow = mysql_fetch_row(xResult);
	if ((sBuf = xRow[0]))
	{
		xClient->m_xPkt_Group.m_nGID = atoi(sBuf);
		if ((sBuf = xRow[1]))
		{
			strncpy(xClient->m_xPkt_Group.m_sGroupname, sBuf, sizeof(xClient->m_xPkt_Group.m_sGroupname)-1);
			if ((sBuf = xRow[2]))
			{
				// We don't return passwords.
				//strncpy(xClient->m_xPkt_Group.m_sPassword, sBuf, sizeof(xClient->m_xPkt_Group.m_sPassword)-1);
				strncpy(xClient->m_xPkt_Group.m_sPassword, "*", sizeof(xClient->m_xPkt_Group.m_sPassword)-1);
				if (MY_ExecA(xClient, &xResult2, g_xConfig[c_nQuery_Group_From_GroupID].m_sValue, xClient->m_xPkt_Group.m_nGID))
				{
					int i2, nMax2 = mysql_num_rows(xResult2);

					xClient->m_xPkt_Group.m_nNumMembers = nMax2;
					xClient->m_xPkt_Answer.m_nAnswer = c_nAnswer_Granted;
					xClient->m_xPkt_Answer.m_nDatatype = c_nDatatype_Group;
					bRet = fuzzd_ReturnAnswer(xClient);
			
					D(DLOG("[%s] MY_ReturnGroup() - nMax2=%d\n", xClient->m_sHostname, nMax2));
					for (i2 = 0; i2 < nMax2; i2++)
					{
						xRow2 = mysql_fetch_row(xResult2);
						sBuf = xRow[0];
						strncpy(xClient->m_xPkt_String.m_sString,	sBuf, sizeof(xClient->m_xPkt_String.m_sString)-1);
						xClient->m_xPkt_Answer.m_nAnswer = c_nAnswer_Granted;
						xClient->m_xPkt_Answer.m_nDatatype = c_nDatatype_String;
						bRet = fuzzd_ReturnAnswer(xClient);
					}
				}
				else
				{
					xClient->m_xPkt_Answer.m_nAnswer = c_nAnswer_Granted;
					xClient->m_xPkt_Answer.m_nDatatype = c_nDatatype_Group;
					bRet = fuzzd_ReturnAnswer(xClient);
				}
				if (xResult2) mysql_free_result(xResult2);
			}
		}
	}
	return bRet;
}

/*****************************
 Get Groups from our backend.
*****************************/
bool MY_GetGroups(struct Client *xClient)
{
	bool		bRet = false;
	MYSQL_RES	*xResult = NULL;

	D(DLOG("[%s] MY_GetUsers()\n", xClient->m_sHostname));
	if (MY_ExecA(xClient, &xResult, g_xConfig[c_nQuery_All_Groups].m_sValue))
	{
		int	i, nMax = mysql_num_rows(xResult);
		D(DLOG("[%s] MY_GetUsers() - Returning %d groups\n", xClient->m_sHostname, nMax));
		bRet = true;
		for (i = 0; (bRet == true) && (i < nMax); i++) bRet = MY_ReturnGroup(xClient, xResult);
	}
	if (xResult) mysql_free_result(xResult);
	// This one will be send by the mainloop.
	xClient->m_xPkt_Answer.m_nAnswer = c_nAnswer_NoMoreData;
	return bRet;
}

/***********************
 Match Group by GroupID
***********************/
bool MY_MatchGroupID(struct Client *xClient)
{
	bool		bRet = false;
	MYSQL_RES	*xResult = NULL;

	D(DLOG("[%s] MY_MatchGroupID(%d)\n", xClient->m_sHostname, xClient->m_xPkt_Number.m_nNumber));
	if (MY_ExecA(xClient, &xResult, g_xConfig[c_nQuery_Group_From_GroupID].m_sValue, xClient->m_xPkt_Number.m_nNumber))
	{
		if (mysql_num_rows(xResult) == 1)
		{
			bRet = MY_ReturnGroup(xClient, xResult);
			// Already Sent...
			xClient->m_xPkt_Answer.m_nAnswer = c_nAnswer_AlreadySent;
		}
	}
	if (xResult) mysql_free_result(xResult);
	if (!bRet) xClient->m_xPkt_Answer.m_nAnswer = c_nAnswer_Denied;
	return bRet;
}

/********************
 Match Group by name
********************/
bool MY_MatchGroupName(struct Client *xClient)
{
	bool		bRet = false;
	MYSQL_RES	*xResult = NULL;
	char		*sGroupname = NULL;

	D(DLOG("[%s] MY_MatchGroupName('%s')\n", xClient->m_sHostname, xClient->m_xPkt_String.m_sString));
	if ((sGroupname = SQLEscape(xClient->m_xPkt_String.m_sString)))
	{
		if (MY_ExecA(xClient, &xResult, g_xConfig[c_nQuery_Group_From_Groupname].m_sValue, sGroupname))
		{
			if (mysql_num_rows(xResult) == 1)
			{
				bRet = MY_ReturnGroup(xClient, xResult);
				// Already Sent...
				xClient->m_xPkt_Answer.m_nAnswer = c_nAnswer_AlreadySent;
			}
		}
		if (xResult) mysql_free_result(xResult);
		free(sGroupname);
	}
	else fuzzd_Log(LOG_ERR, "[%s] Couldn't allocate memory for escaping groupname\n", xClient->m_sHostname);
	if (!bRet) xClient->m_xPkt_Answer.m_nAnswer = c_nAnswer_Denied;
	return bRet;
}

/********************
 Add a new User
********************/
bool MY_UserAdd(struct Client *xClient)
{
	bool bRet = false;
	char *sCrypt;

	D(DLOG("[%s] MY_UserAdd()\n", xClient->m_sHostname));
	if ((sCrypt = fuzzd_Encrypt(xClient->m_xPkt_User.m_sPasswordNew)))
	{
		char		*sUsername = SQLEscape(xClient->m_xPkt_User.m_sUsername);
		if (sUsername)
		{
			char *sCryptSQL = SQLEscape(sCrypt);
			if (sCryptSQL)
			{
				MYSQL_RES *xResult = NULL;
				// Add the user to the database.
				if (MY_ExecA(xClient, &xResult, g_xConfig[c_nQuery_UserAdd_User].m_sValue, sUsername, xClient->m_xPkt_User.m_nGID, xClient->m_xPkt_User.m_sGECOS, xClient->m_xPkt_User.m_sHomedir, xClient->m_xPkt_User.m_sShell))
				{
					MYSQL_RES *xResult2 = NULL;
					MYSQL_ROW xRow2;
					// Fetch the users UID.
					if (MY_ExecA(xClient, &xResult2, g_xConfig[c_nQuery_User_From_Username].m_sValue, sUsername))
					{
						char *sUID;
						// We actually fetch everything and only need user_uid.
						if ((mysql_num_rows(xResult2) == 7) && (xRow2 = mysql_fetch_row(xResult2)))
						{
							MYSQL_RES *xResult3 = NULL;
							sUID = xRow2[0];
							xClient->m_xPkt_User.m_nUID = atoi(sUID);
							// Add the user to it's primary group.
							if (MY_ExecA(xClient, &xResult3, g_xConfig[c_nQuery_UserAdd_Group].m_sValue, xClient->m_xPkt_User.m_nGID, xClient->m_xPkt_User.m_nUID))
							{
								MYSQL_RES *xResult4 = NULL;
								// Add the user's password to the shadowlist.
								if (MY_ExecA(xClient, &xResult4, g_xConfig[c_nQuery_UserAdd_Password].m_sValue, xClient->m_xPkt_User.m_nUID, sCryptSQL))
								{
									fuzzd_Log(LOG_INFO, "[%s] User %s added succesfully.\n", xClient->m_sHostname, xClient->m_xPkt_User.m_sUsername);
									
									// Call the Script.
									bRet = true;
								}
								else fuzzd_Log(LOG_ERR, "[%s] User %s added, but couldn't add it's password: %s.\n", xClient->m_sHostname, xClient->m_xPkt_User.m_sUsername, mysql_error(&g_xConn));
								if (xResult4) mysql_free_result(xResult4);
							}
							else fuzzd_Log(LOG_ERR, "[%s] User %s added, but couldn't add it to the primary group %d: %s.\n", xClient->m_sHostname, xClient->m_xPkt_User.m_sUsername, xClient->m_xPkt_User.m_nGID, mysql_error(&g_xConn));
							if (xResult3) mysql_free_result(xResult3);
						}
						else fuzzd_Log(LOG_ERR, "[%s] User %s added, but couldn't find it's user id back: %s.\n", xClient->m_sHostname, xClient->m_xPkt_User.m_sUsername, mysql_error(&g_xConn));
					}
					if (xResult2) mysql_free_result(xResult2);
					bRet = true;
				}
				else fuzzd_Log(LOG_ERR, "[%s] Adding user %s failed: %s.\n", xClient->m_sHostname, xClient->m_xPkt_User.m_sUsername, mysql_error(&g_xConn));
				if (xResult) mysql_free_result(xResult);
				free(sCryptSQL);
			}
			else fuzzd_Log(LOG_ERR, "[%s] Couldn't allocate memory for escaping password\n", xClient->m_sHostname);
			free(sUsername);
		}
		else fuzzd_Log(LOG_ERR, "[%s] Couldn't allocate memory for escaping username\n", xClient->m_sHostname);
	}
	return bRet;
}

/********************
 Remove a User
********************/
bool MY_UserRemove(struct Client *xClient)
{
	bool	bRet = false;
	char	*sUsername = SQLEscape(xClient->m_xPkt_String.m_sString);
	D(DLOG("[%s] MY_UserRemove()\n", xClient->m_sHostname));
	if (sUsername)
	{
		MYSQL_RES *xResult = NULL;
		// Add the user to the database.
		if (MY_ExecA(xClient, &xResult, g_xConfig[c_nQuery_User_From_Username].m_sValue, sUsername))
		{
			MYSQL_ROW	xRow;
			// We actually fetch everything and only need user_uid.
			if ((mysql_num_rows(xResult) == 7) && (xRow = mysql_fetch_row(xResult)))
			{
				int nUID = atoi(xRow[0]);
				MYSQL_RES *xResult2 = NULL;

				// Remove the passwordlist (shadow).
				if (MY_ExecA(xClient, &xResult2, g_xConfig[c_nQuery_UserRemove_Password].m_sValue, nUID))
				{
					MYSQL_RES *xResult3 = NULL;
					// Remove the user from all the groups.
					if (MY_ExecA(xClient, &xResult3, g_xConfig[c_nQuery_UserRemove_Group].m_sValue, nUID))
					{
						MYSQL_RES *xResult4 = NULL;
						// Remove the user from the userlist.
						if (MY_ExecA(xClient, &xResult4, g_xConfig[c_nQuery_UserRemove_User].m_sValue, nUID))
						{
							fuzzd_Log(LOG_INFO, "[%s] User %s removed succesfully.\n", xClient->m_sHostname, xClient->m_xPkt_String.m_sString);
							
							// Call the cleanup script.
							bRet = true;
						}
						else fuzzd_Log(LOG_ERR, "[%s] User %s partially removed, couldn't remove user entry: %s.\n", xClient->m_sHostname, xClient->m_xPkt_String.m_sString, mysql_error(&g_xConn));
						if (xResult4) mysql_free_result(xResult4);
					}
					else fuzzd_Log(LOG_ERR, "[%s] User %s partially removed, couldn't remove user from any groups: %s.\n", xClient->m_sHostname, xClient->m_xPkt_String.m_sString, mysql_error(&g_xConn));
					if (xResult3) mysql_free_result(xResult3);
				}
				else fuzzd_Log(LOG_ERR, "[%s] User %s partially removed, couldn't remove user from password (shadow): %s.\n", xClient->m_sHostname, xClient->m_xPkt_String.m_sString, mysql_error(&g_xConn));
				if (xResult2) mysql_free_result(xResult2);
			}
		}
		else fuzzd_Log(LOG_ERR, "[%s] Removing user %s failed: %s.\n", xClient->m_sHostname, xClient->m_xPkt_String.m_sString, mysql_error(&g_xConn));
		if (xResult) mysql_free_result(xResult);
		free(sUsername);
	}
	else fuzzd_Log(LOG_ERR, "[%s] Couldn't allocate memory for escaping username\n", xClient->m_sHostname);
	return bRet;
}

/********************
 Change a User
********************/
bool MY_UserChange(struct Client *xClient)
{
	bool bRet = false;
	return bRet;
	xClient = xClient;
}

/********************
 Add a new Group
********************/
bool MY_GroupAdd(struct Client *xClient)
{
	bool bRet = false;
	return bRet;
	xClient = xClient;
}

/********************
 Remove a Group
********************/
bool MY_GroupRemove(struct Client *xClient)
{
	bool bRet = false;
	return bRet;
	xClient = xClient;
}

/********************
 Change a Group
********************/
bool MY_GroupChange(struct Client *xClient)
{
	bool bRet = false;
	return bRet;
	xClient = xClient;
}

/********************
 Add a new Member
********************/
bool MY_MemberAdd(struct Client *xClient)
{
	bool bRet = false;
	return bRet;
	xClient = xClient;
}

/********************
 Remove a Member
********************/
bool MY_MemberRemove(struct Client *xClient)
{
	bool bRet = false;
	return bRet;
	xClient = xClient;
}

/**************
 Authenticator
**************/
struct Authenticator mysql_Authenticator = {
	MY_Initialize,
	MY_Shutdown,
	MY_GetInfo,
	MY_Authenticate,
	MY_ChangePassword,
	MY_GetUsers,
	MY_MatchUserID,
	MY_MatchUserName,
	MY_MatchUserGECOS,
	MY_GetGroups,	
	MY_MatchGroupID,
	MY_MatchGroupName,
	MY_UserAdd,	
	MY_UserRemove,
	MY_UserChange,
	MY_GroupAdd,
	MY_GroupRemove,
	MY_GroupChange,
	MY_MemberAdd,
	MY_MemberRemove
};
