/* FUZZd PostgreSQL Authenticator by Jeroen Massar <jeroen@unfix.org> */

// FUZZd
#include "fuzzd.h"
#include "fuzzauth_config.h"

// Postgres Includes
#include <postgresql/libpq-fe.h>

// Globals
PGconn *g_xConn;

/*****************************
 PostgreSQL notice & warnings
*****************************/
static void PG_NoticeProcessor(void *sArg, const char *sMessage)
{
	fuzzd_Log(LOG_INFO, "PostgreSQL notice/warning: %s\n", sMessage);
	return;
	sArg = sArg;
}

struct Config g_xConfig[] ={
										{"Database",								NULL, 0},
										{"Servername",								NULL, 0},
										{"Username",								NULL, 0},
										{"Password",								NULL, 0},
										// Queries
										{"Query_Password_From_User",			NULL, 0},
										{"Query_Update_Password",				NULL, 0},
										{"Query_All_Users",						NULL, 0},
										{"Query_User_From_UserID",				NULL, 0},
										{"Query_User_From_Username",			NULL, 0},
										{"Query_User_From_GECOS",				NULL, 0},
										{"Query_GroupMember_From_GroupID",	NULL, 0},
										{"Query_All_Groups",						NULL, 0},
										{"Query_Group_From_GroupID",			NULL, 0},
										{"Query_Group_From_Groupname",		NULL, 0},
										// UserAdd()
										{"UserAdd_Script",						NULL, 0},
										{"Query_UserAdd_User",					NULL, 0},
										{"Query_UserAdd_Group",					NULL, 0},
										{"Query_UserAdd_Password",				NULL, 0},
										// UserRemove()
										{"UserRemove_Script",					NULL, 0},
										{"Query_UserRemove_User",				NULL, 0},
										{"Query_UserRemove_Group",				NULL, 0},
										{"Query_UserRemove_Password",			NULL, 0}
									};

enum c_nConfig {
										c_nDatabase = 0,
										c_nServername,
										c_nUsername,
										c_nPassword,
										// Queries
										c_nQuery_Password_From_User,
										c_nQuery_Update_Password,
										c_nQuery_All_Users,
										c_nQuery_User_From_UserID,
										c_nQuery_User_From_Username,
										c_nQuery_User_From_GECOS,
										c_nQuery_GroupMember_From_GroupID,
										c_nQuery_All_Groups,
										c_nQuery_Group_From_GroupID,
										c_nQuery_Group_From_Groupname,
										// UserAdd()
										c_nUserAdd_Script,
										c_nQuery_UserAdd_User,
										c_nQuery_UserAdd_UserID,
										c_nQuery_UserAdd_Group,
										c_nQuery_UserAdd_Password,
										c_nUserRemove_Script,
										c_nQuery_UserRemove_User,
										c_nQuery_UserRemove_Group,
										c_nQuery_UserRemove_Password
					};

/***********************
 Initialize the module
***********************/
bool pg_Initialize(void)
{
	// NOTE: We will want to add 'requiressl' here if the host != localhost.
	//       Then again, one should probably run a fuzzd on the database host ;)
	bool				bRet = false;
	unsigned int	i;
	char				*sError = NULL;

	D(DLOG("pg_Initialize()\n"));
	sError = ParseConfig("auth_postgresql", g_xConfig, sizeof(g_xConfig)/sizeof(struct Config));
	if (sError == NULL)
	{
		for (bRet = true, i = 0; bRet && (i<sizeof(g_xConfig)/sizeof(struct Config)); i++)
		{
//			D(DLOG("Checking %d: %s -> %s\n", i, g_xConfig[i].m_sName, g_xConfig[i].m_sValue));
			if (g_xConfig[i].m_sValue == NULL) bRet = false;
		}
		if (bRet)
		{
			char *sConnectString;
			asprintf(&sConnectString, "dbname=%s host=%s user=%s password=%s", g_xConfig[c_nDatabase].m_sValue, g_xConfig[c_nServername].m_sValue, g_xConfig[c_nUsername].m_sValue, g_xConfig[c_nPassword].m_sValue);
			if (sConnectString)
			{
				g_xConn = PQconnectdb(sConnectString);
				if (PQstatus(g_xConn) == CONNECTION_OK)
				{
					fuzzd_Log(LOG_INFO, "PostgreSQL connection succesful.\n");
					PQsetNoticeProcessor(g_xConn, PG_NoticeProcessor, NULL);
					bRet = true;
				}
				else
				{
					fuzzd_Log(LOG_ERR, "PostgreSQL connection failed: '%s'\n", PQerrorMessage(g_xConn));
					bRet = false;
				}
				free(sConnectString);
			}
		}
		else
		{
			i--;
			fuzzd_Log(LOG_ERR, "Unset option ('%s') detected ...\n", g_xConfig[i].m_sName);
		}
	}
	else
	{
		fuzzd_Log(LOG_ERR, "%s", sError);
		free(sError);
	}
	return bRet;
}

/*********************
 Shutdown the module
*********************/
bool pg_Shutdown(void)
{
//	D(DLOG("pg_Shutdown()\n"));
	if (g_xConn)
	{
		PQfinish(g_xConn);
		g_xConn = NULL;
	}
	return true;
}

/******************************
 Get some information.
******************************/
bool pg_GetInfo(char *sBuf, int nSize)
{
	strncpy(sBuf, "PostgreSQL FUZZauth Authenticator Plugin by Jeroen Massar\n", nSize);
	return true;
}

/******************************
 (local) Execute an SQL Query.
******************************/
bool PG_Exec(struct Client *xClient, PGresult **xResult, char *sQuery)
{
	bool bRet = false;

	D(DLOG("[%s] pg_Exec(%s)\n", xClient->m_sHostname, sQuery));
	*xResult = PQexec(g_xConn, sQuery);
	if ((PQresultStatus(*xResult) == PGRES_COMMAND_OK) ||
		 (PQresultStatus(*xResult) == PGRES_TUPLES_OK))
	{
		//D(DLOG("[%s] PostgreSQL query succesful: '%s'\n", xClient->m_sHostname, PQresultErrorMessage(*xResult)));
		bRet = true;
	}
	else
	{
		fuzzd_Log(LOG_ERR, "[%s] PostgreSQL query failed: '%s'\n", xClient->m_sHostname, PQresultErrorMessage(*xResult));
		bRet = false;
	}
	return bRet;
}

/******************************************
 (local) Execute an SQL Query (Formatted).
******************************************/
bool PG_ExecA(struct Client *xClient, PGresult **xResult, char *sFormat, ...)
{
	bool		bRet = false;
	int		nLen;
	char		*sQuery = NULL;
	va_list	xArgs;

	va_start(xArgs, sFormat);
	nLen = vasprintf(&sQuery, sFormat, xArgs);
	if (sQuery)
	{
		bRet = PG_Exec(xClient, xResult, sQuery);
		free(sQuery);
	}
	va_end(xArgs);
	return bRet;
}

/**************************
 Escape all SQL-escapes.
 the result must be free()'d by the caller.
**************************/
char *SQLEscape(char *sString)
{
	int	i = 0, j = 0, nEscapes = 0, nNewLen = 0;
	char	*sEscaped = NULL;

	while (sString[i] != '\0')
	{
		if (	(sString[i] == '\'') ||
				(sString[i] == '\"') ||
				(sString[i] == '\\'))
		{
			nEscapes++;
		}
		i++;
	}

	nNewLen = i + nEscapes;
//	D(DLOG("SQLEscape() - Old = %d/%d, New = %d\n", strlen(sString), i, nNewLen));

	if ((sEscaped = malloc(nNewLen + 1)))
	{
		i = 0;
		while (sString[i] != '\0')
		{
			if (	(sString[i] == '\'') ||
					(sString[i] == '\"') ||
					(sString[i] == '\\'))
			{
				sEscaped[j] = '\\';
				j++;
				sEscaped[j] = sString[i];
			}
			else sEscaped[j] = sString[i];
			j++;
			i++;
		}
		// Extra verification.
		if (j == (nNewLen))
		{
			sEscaped[j] = '\0';
//			D(DLOG("SQLEscape() - \"%s\" (%d) -> \"%s\" (%d/%d)\n", sString, strlen(sString), sEscaped, strlen(sEscaped), nNewLen));
		}
		else
		{
//			D(DLOG("SQLEscape() - Process failed (j=%d, nNewLen=%d).\n", j, nNewLen));
			free(sEscaped);
			sEscaped = NULL;
		}
	}
	return sEscaped;
}

/**************************
 Authenticate this Client
**************************/
bool pg_Authenticate(struct Client *xClient)
{
	bool			bRet = false;
	PGresult		*xResult = NULL;
	char			*sUsername = NULL;

//	D(DLOG("[%s] pg_Authenticate()\n", xClient->m_sHostname));
	if ((sUsername = SQLEscape(xClient->m_xPkt_User.m_sUsername)))
	{
		if (PG_ExecA(xClient, &xResult, g_xConfig[c_nQuery_Password_From_User].m_sValue, sUsername))
		{
			if (PQntuples(xResult) == 1)
			{
				char *sPassword = PQgetvalue(xResult, 0, 0);
				if (sPassword)
				{
					// If we are being called by ChangePassword, also allow setting a password if it's empty...
					// This allows an administrator to clear the password in the database and then quickly update it.
//					D(DLOG("[%s] pg_Authenticate() - Password Length = %d, Cmd=%d\n", xClient->m_sHostname, strlen(sPassword), xClient->m_xPkt_Request.m_nCommand));
					if ((xClient->m_xPkt_Request.m_nCommand == c_nCommand_ChangePassword) && (sPassword) && (strlen(sPassword) == 0))
					{
//						D(DLOG("[%s] pg_Authenticate() - Succeeded because of ChangePassword and the database password is empty.\n", xClient->m_sHostname));
						bRet = true;
					}
					// And in any case try to compare the passwords like usual.
					else if ((sPassword) && (fuzzd_ComparePassword(sPassword, xClient->m_xPkt_User.m_sPassword)))
					{
//						D(DLOG("[%s] pg_Authenticate() - Passwords match.\n", xClient->m_sHostname));
						bRet = true;
					}
					else fuzzd_Log(LOG_ERR, "[%s] pg_Authenticate() - Passwords MISmatch.\n", xClient->m_sHostname);
				}
//				D(else DLOG("[%s] pg_Authenticate() - No password in database.\n", xClient->m_sHostname));
			}
			else
			{
//				D(DLOG("[%s] pg_Authenticate() - No tuples returned.\n", xClient->m_sHostname));
				bRet = false;
			}
		}
		free(sUsername);
	}
	else fuzzd_Log(LOG_ERR, "[%s] Couldn't allocate memory for escaping username\n", xClient->m_sHostname);
	if (xResult) PQclear(xResult);
	return bRet;
}

/****************
 Change Password
****************/
bool pg_ChangePassword(struct Client *xClient)
{
	bool bRet = false;
	char *sCrypt;

//	D(DLOG("[%s] pg_ChangePassword()\n", xClient->m_sHostname));
	if ((sCrypt = fuzzd_Encrypt(xClient->m_xPkt_User.m_sPasswordNew)))
	{
		PGresult *xResult = NULL;
		char		*sUsername = SQLEscape(xClient->m_xPkt_User.m_sUsername);
		if (sUsername)
		{
			char *sCryptSQL = SQLEscape(sCrypt);
			if (sCryptSQL)
			{
				if (PG_ExecA(xClient, &xResult, g_xConfig[c_nQuery_Update_Password].m_sValue, sCryptSQL, sUsername))
				{
					fuzzd_Log(LOG_INFO, "[%s] Password for user %s changed succesfully.\n", xClient->m_sHostname, xClient->m_xPkt_User.m_sUsername);
					bRet = true;
				}
				else fuzzd_Log(LOG_ERR, "[%s] Password changing for user %s failed: %s.\n", xClient->m_sHostname, xClient->m_xPkt_User.m_sUsername, PQerrorMessage(g_xConn));
				free(sCryptSQL);
			}
			else fuzzd_Log(LOG_ERR, "[%s] Couldn't allocate memory for escaping password\n", xClient->m_sHostname);
			free(sUsername);
		}
		else fuzzd_Log(LOG_ERR, "[%s] Couldn't allocate memory for escaping username\n", xClient->m_sHostname);
	}
	return bRet;
}

/********************************
 Return a User from SQL results.
********************************/
bool PG_ReturnUser(struct Client *xClient, PGresult *xResult, int i)
{
	bool bRet = false;
	char *sBuf;

//	D(DLOG("[%s] PG_ReturnUser()\n", xClient->m_sHostname));
	// Clear the complete thing.
	memset(&xClient->m_xPkt_User, 0, sizeof(xClient->m_xPkt_User));
	if ((sBuf = PQgetvalue(xResult, i, 0)))
	{
		strncpy(xClient->m_xPkt_User.m_sUsername, sBuf, sizeof(xClient->m_xPkt_User.m_sUsername)-1);
		if ((sBuf = PQgetvalue(xResult, i, 1)))
		{
			// We don't return passwords.
			//strncpy(xClient->m_xPkt_User.m_sPassword, sBuf, sizeof(xClient->m_xPkt_User.m_sPassword)-1);
			strncpy(xClient->m_xPkt_User.m_sPassword, "*", sizeof(xClient->m_xPkt_User.m_sPassword)-1);
			if ((sBuf = PQgetvalue(xResult, i, 2)))
			{
				xClient->m_xPkt_User.m_nUID = atoi(sBuf);
				if ((xClient->m_xPkt_User.m_nUID > 1000) && (sBuf = PQgetvalue(xResult, i, 3)))
				{
					xClient->m_xPkt_User.m_nGID = atoi(sBuf);
					if ((xClient->m_xPkt_User.m_nGID > 1000) && (sBuf = PQgetvalue(xResult, i, 4)))
					{
						strncpy(xClient->m_xPkt_User.m_sGECOS,	sBuf, sizeof(xClient->m_xPkt_User.m_sGECOS)-1);
						if ((sBuf = PQgetvalue(xResult, i, 5)))
						{
							strncpy(xClient->m_xPkt_User.m_sHomedir, sBuf, sizeof(xClient->m_xPkt_User.m_sHomedir)-1);
							if ((sBuf = PQgetvalue(xResult, i, 6)))
							{
								D(DLOG("[%s] PG_ReturnUser() - %d:%s:%s\n", xClient->m_sHostname, xClient->m_xPkt_User.m_nUID, xClient->m_xPkt_User.m_sUsername, xClient->m_xPkt_User.m_sGECOS));
								strncpy(xClient->m_xPkt_User.m_sShell,	sBuf, sizeof(xClient->m_xPkt_User.m_sShell)-1);
								xClient->m_xPkt_Answer.m_nAnswer = c_nAnswer_Granted;
								xClient->m_xPkt_Answer.m_nDatatype = c_nDatatype_User;
								bRet = fuzzd_ReturnAnswer(xClient);
							}
//							D(else DLOG("[%s] pg_GetUsers() - No value #7\n", xClient->m_sHostname));
						}
//						D(else DLOG("[%s] pg_GetUsers() - No value #6\n", xClient->m_sHostname));
					}
//					D(else DLOG("[%s] pg_GetUsers() - No value #5\n", xClient->m_sHostname));
				}
//				D(else DLOG("[%s] pg_GetUsers() - No value #4\n", xClient->m_sHostname));
			}
//			D(else DLOG("[%s] pg_GetUsers() - No value #3\n", xClient->m_sHostname));
		}
//		D(else DLOG("[%s] pg_GetUsers() - No value #2\n", xClient->m_sHostname));
	}
//	D(else DLOG("[%s] pg_GetUsers() - No value #1\n", xClient->m_sHostname));
	return bRet;
}

/****************************
 Get Users from our backend.
****************************/
bool pg_GetUsers(struct Client *xClient)
{
	bool		bRet = false;
	PGresult	*xResult = NULL;

//	D(DLOG("[%s] pg_GetUsers()\n", xClient->m_sHostname));
	if (PG_ExecA(xClient, &xResult, g_xConfig[c_nQuery_All_Users].m_sValue))
	{
		int	i, nMax = PQntuples(xResult);

//		D(DLOG("[%s] pg_GetUsers() - Returning %d users\n", xClient->m_sHostname, nMax));
		bRet = true;
		for (i = 0; (bRet == true) && (i < nMax); i++) bRet = PG_ReturnUser(xClient, xResult, i);
	}
	if (xResult) PQclear(xResult);
	// This one will be send by the mainloop.
	xClient->m_xPkt_Answer.m_nAnswer = c_nAnswer_NoMoreData;
	return bRet;
}

/*********************
 Match User by UserID
*********************/
bool pg_MatchUserID(struct Client *xClient)
{
	bool		bRet = false;
	PGresult	*xResult = NULL;

//	D(DLOG("[%s] pg_MatchUserID(%d)\n", xClient->m_sHostname, xClient->m_xPkt_Number.m_nNumber));
	if (PG_ExecA(xClient, &xResult, g_xConfig[c_nQuery_User_From_UserID].m_sValue, xClient->m_xPkt_Number.m_nNumber))
	{
		if (PQntuples(xResult) == 1)
		{
			bRet = PG_ReturnUser(xClient, xResult, 0);
			// Already Sent...
			xClient->m_xPkt_Answer.m_nAnswer = c_nAnswer_AlreadySent;
		}
	}
	if (xResult) PQclear(xResult);
	if (!bRet) xClient->m_xPkt_Answer.m_nAnswer = c_nAnswer_Denied;
	return bRet;
}

/*******************
 Match User by name
*******************/
bool pg_MatchUserName(struct Client *xClient)
{
	bool		bRet = false;
	PGresult	*xResult = NULL;
	char		*sUsername = NULL;

//	D(DLOG("[%s] pg_MatchUserName('%s')\n", xClient->m_sHostname, xClient->m_xPkt_String.m_sString));
	if ((sUsername = SQLEscape(xClient->m_xPkt_String.m_sString)))
	{
		if (PG_ExecA(xClient, &xResult, g_xConfig[c_nQuery_User_From_Username].m_sValue, sUsername))
		{
			if (PQntuples(xResult) == 1)
			{
				bRet = PG_ReturnUser(xClient, xResult, 0);
				// Already Sent...
				xClient->m_xPkt_Answer.m_nAnswer = c_nAnswer_AlreadySent;
			}
		}
		if (xResult) PQclear(xResult);
		free(sUsername);
	}
//	else fuzzd_Log(LOG_ERR, "[%s] Couldn't allocate memory for escaping username\n", xClient->m_sHostname);
	if (!bRet) xClient->m_xPkt_Answer.m_nAnswer = c_nAnswer_Denied;
	return bRet;
}

/********************
 Match User by GECOS
********************/
bool pg_MatchUserGECOS(struct Client *xClient)
{
	bool		bRet = false;
	PGresult	*xResult = NULL;
	char		*sGECOS = NULL;

//	D(DLOG("[%s] pg_MatchUserGECOS('%s')\n", xClient->m_sHostname, xClient->m_xPkt_String.m_sString));
	if ((sGECOS = SQLEscape(xClient->m_xPkt_String.m_sString)))
	{
		if (PG_ExecA(xClient, &xResult, g_xConfig[c_nQuery_User_From_GECOS].m_sValue, sGECOS))
		{
			if (PQntuples(xResult) == 1)
			{
				bRet = PG_ReturnUser(xClient, xResult, 0);
				// Already Sent...
				xClient->m_xPkt_Answer.m_nAnswer = c_nAnswer_AlreadySent;
			}
		}
		if (xResult) PQclear(xResult);
		free(sGECOS);
	}
//	else fuzzd_Log(LOG_ERR, "[%s] Couldn't allocate memory for escaping username\n", xClient->m_sHostname);
	if (!bRet) xClient->m_xPkt_Answer.m_nAnswer = c_nAnswer_Denied;
	return bRet;
}

/********************************
 Return a Group from SQL results.
********************************/
bool PG_ReturnGroup(struct Client *xClient, PGresult *xResult, int i)
{
	bool		bRet = false;
	char		*sBuf;
	PGresult	*xResult2 = NULL;

	D(DLOG("[%s] PG_ReturnGroup(%d)\n", xClient->m_sHostname, i));
	// Clear the complete thing.
	memset(&xClient->m_xPkt_Group, 0, sizeof(xClient->m_xPkt_Group));
	if ((sBuf = PQgetvalue(xResult, i, 0)))
	{
		xClient->m_xPkt_Group.m_nGID = atoi(sBuf);
		if ((sBuf = PQgetvalue(xResult, i, 1)))
		{
			strncpy(xClient->m_xPkt_Group.m_sGroupname, sBuf, sizeof(xClient->m_xPkt_Group.m_sGroupname)-1);
			if ((sBuf = PQgetvalue(xResult, i, 2)))
			{
				// We don't return passwords.
				//strncpy(xClient->m_xPkt_Group.m_sPassword, sBuf, sizeof(xClient->m_xPkt_Group.m_sPassword)-1);
				strncpy(xClient->m_xPkt_Group.m_sPassword, "*", sizeof(xClient->m_xPkt_Group.m_sPassword)-1);
				if (PG_ExecA(xClient, &xResult2, g_xConfig[c_nQuery_GroupMember_From_GroupID].m_sValue, xClient->m_xPkt_Group.m_nGID))
				{
					int i2, nMax2 = PQntuples(xResult2);

					xClient->m_xPkt_Group.m_nNumMembers = nMax2;
					xClient->m_xPkt_Answer.m_nAnswer = c_nAnswer_Granted;
					xClient->m_xPkt_Answer.m_nDatatype = c_nDatatype_Group;
					bRet = fuzzd_ReturnAnswer(xClient);
			
					D(DLOG("[%s] PG_ReturnGroup(%d) - nMax2=%d\n", xClient->m_sHostname, i, nMax2));
					for (i2 = 0; i2 < nMax2; i2++)
					{
						sBuf = PQgetvalue(xResult2, i2, 0);
						strncpy(xClient->m_xPkt_String.m_sString,	sBuf, sizeof(xClient->m_xPkt_String.m_sString)-1);
						xClient->m_xPkt_Answer.m_nAnswer = c_nAnswer_Granted;
						xClient->m_xPkt_Answer.m_nDatatype = c_nDatatype_String;
						bRet = fuzzd_ReturnAnswer(xClient);
					}
				}
				else
				{
					xClient->m_xPkt_Answer.m_nAnswer = c_nAnswer_Granted;
					xClient->m_xPkt_Answer.m_nDatatype = c_nDatatype_Group;
					bRet = fuzzd_ReturnAnswer(xClient);
				}
				if (xResult2) PQclear(xResult2);
			}
		}
	}
	return bRet;
}

/*****************************
 Get Groups from our backend.
*****************************/
bool pg_GetGroups(struct Client *xClient)
{
	bool		bRet = false;
	PGresult	*xResult = NULL;

	D(DLOG("[%s] pg_GetUsers()\n", xClient->m_sHostname));
	if (PG_ExecA(xClient, &xResult, g_xConfig[c_nQuery_All_Groups].m_sValue))
	{
		int	i, nMax = PQntuples(xResult);
		D(DLOG("[%s] pg_GetGroups() - Returning %d groups\n", xClient->m_sHostname, nMax));
		bRet = true;
		for (i = 0; (bRet == true) && (i < nMax); i++) bRet = PG_ReturnGroup(xClient, xResult, i);
	}
	if (xResult) PQclear(xResult);
	// This one will be send by the mainloop.
	xClient->m_xPkt_Answer.m_nAnswer = c_nAnswer_NoMoreData;
	return bRet;
}

/***********************
 Match Group by GroupID
***********************/
bool pg_MatchGroupID(struct Client *xClient)
{
	bool		bRet = false;
	PGresult	*xResult = NULL;

	D(DLOG("[%s] pg_MatchGroupID(%d)\n", xClient->m_sHostname, xClient->m_xPkt_Number.m_nNumber));
	if (PG_ExecA(xClient, &xResult, g_xConfig[c_nQuery_Group_From_GroupID].m_sValue, xClient->m_xPkt_Number.m_nNumber))
	{
		if (PQntuples(xResult) == 1)
		{
			bRet = PG_ReturnGroup(xClient, xResult, 0);
			// Already Sent...
			xClient->m_xPkt_Answer.m_nAnswer = c_nAnswer_AlreadySent;
		}
	}
	if (xResult) PQclear(xResult);
	if (!bRet) xClient->m_xPkt_Answer.m_nAnswer = c_nAnswer_Denied;
	return bRet;
}

/********************
 Match Group by name
********************/
bool pg_MatchGroupName(struct Client *xClient)
{
	bool		bRet = false;
	PGresult	*xResult = NULL;
	char		*sGroupname = NULL;

	D(DLOG("[%s] pg_MatchGroupName('%s')\n", xClient->m_sHostname, xClient->m_xPkt_String.m_sString));
	if ((sGroupname = SQLEscape(xClient->m_xPkt_String.m_sString)))
	{
		if (PG_ExecA(xClient, &xResult, g_xConfig[c_nQuery_Group_From_Groupname].m_sValue, sGroupname))
		{
			if (PQntuples(xResult) == 1)
			{
				bRet = PG_ReturnGroup(xClient, xResult, 0);
				// Already Sent...
				xClient->m_xPkt_Answer.m_nAnswer = c_nAnswer_AlreadySent;
			}
		}
		if (xResult) PQclear(xResult);
		free(sGroupname);
	}
	else fuzzd_Log(LOG_ERR, "[%s] Couldn't allocate memory for escaping groupname\n", xClient->m_sHostname);
	if (!bRet) xClient->m_xPkt_Answer.m_nAnswer = c_nAnswer_Denied;
	return bRet;
}

/********************
 Add a new User
********************/
bool pg_UserAdd(struct Client *xClient)
{
	bool bRet = false;
	char *sCrypt;

	D(DLOG("[%s] pg_UserAdd()\n", xClient->m_sHostname));
	if ((sCrypt = fuzzd_Encrypt(xClient->m_xPkt_User.m_sPasswordNew)))
	{
		char		*sUsername = SQLEscape(xClient->m_xPkt_User.m_sUsername);
		if (sUsername)
		{
			char *sCryptSQL = SQLEscape(sCrypt);
			if (sCryptSQL)
			{
				PGresult *xResult = NULL;
				// Add the user to the database.
				if (PG_ExecA(xClient, &xResult, g_xConfig[c_nQuery_UserAdd_User].m_sValue, sUsername, xClient->m_xPkt_User.m_nGID, xClient->m_xPkt_User.m_sGECOS, xClient->m_xPkt_User.m_sHomedir, xClient->m_xPkt_User.m_sShell))
				{
					PGresult *xResult2 = NULL;
					// Fetch the users UID.
					if (PG_ExecA(xClient, &xResult2, g_xConfig[c_nQuery_User_From_Username].m_sValue, sUsername))
					{
						char *sUID;
						// We actually fetch everything and only need user_uid.
						if ((PQntuples(xResult2) == 7) && (sUID = PQgetvalue(xResult2, 0, 2)))
						{
							PGresult *xResult3 = NULL;
							xClient->m_xPkt_User.m_nUID = atoi(sUID);
							// Add the user to it's primary group.
							if (PG_ExecA(xClient, &xResult3, g_xConfig[c_nQuery_UserAdd_Group].m_sValue, xClient->m_xPkt_User.m_nGID, xClient->m_xPkt_User.m_nUID))
							{
								PGresult *xResult4 = NULL;
								// Add the user's password to the shadowlist.
								if (PG_ExecA(xClient, &xResult4, g_xConfig[c_nQuery_UserAdd_Password].m_sValue, xClient->m_xPkt_User.m_nUID, sCryptSQL))
								{
									fuzzd_Log(LOG_INFO, "[%s] User %s added succesfully.\n", xClient->m_sHostname, xClient->m_xPkt_User.m_sUsername);
									
									// Call the Script.
									bRet = true;
								}
								else fuzzd_Log(LOG_ERR, "[%s] User %s added, but couldn't add it's password: %s.\n", xClient->m_sHostname, xClient->m_xPkt_User.m_sUsername, PQerrorMessage(g_xConn));
								if (xResult4) PQclear(xResult4);
							}
							else fuzzd_Log(LOG_ERR, "[%s] User %s added, but couldn't add it to the primary group %d: %s.\n", xClient->m_sHostname, xClient->m_xPkt_User.m_sUsername, xClient->m_xPkt_User.m_nGID, PQerrorMessage(g_xConn));
							if (xResult3) PQclear(xResult3);
						}
						else fuzzd_Log(LOG_ERR, "[%s] User %s added, but couldn't find it's user id back: %s.\n", xClient->m_sHostname, xClient->m_xPkt_User.m_sUsername, PQerrorMessage(g_xConn));
					}
					if (xResult2) PQclear(xResult2);
					bRet = true;
				}
				else fuzzd_Log(LOG_ERR, "[%s] Adding user %s failed: %s.\n", xClient->m_sHostname, xClient->m_xPkt_User.m_sUsername, PQerrorMessage(g_xConn));
				if (xResult) PQclear(xResult);
				free(sCryptSQL);
			}
			else fuzzd_Log(LOG_ERR, "[%s] Couldn't allocate memory for escaping password\n", xClient->m_sHostname);
			free(sUsername);
		}
		else fuzzd_Log(LOG_ERR, "[%s] Couldn't allocate memory for escaping username\n", xClient->m_sHostname);
	}
	return bRet;
}

/********************
 Remove a User
********************/
bool pg_UserRemove(struct Client *xClient)
{
	bool	bRet = false;
	char	*sUsername = SQLEscape(xClient->m_xPkt_String.m_sString);
	D(DLOG("[%s] pg_UserRemove()\n", xClient->m_sHostname));
	if (sUsername)
	{
		PGresult *xResult = NULL;
		// Add the user to the database.
		if (PG_ExecA(xClient, &xResult, g_xConfig[c_nQuery_User_From_Username].m_sValue, sUsername))
		{
			char *sUID;
			// We actually fetch everything and only need user_uid.
			if ((PQntuples(xResult) == 7) && (sUID = PQgetvalue(xResult, 0, 2)))
			{
				int nUID = atoi(sUID);
				PGresult *xResult2 = NULL;

				// Remove the passwordlist (shadow).
				if (PG_ExecA(xClient, &xResult2, g_xConfig[c_nQuery_UserRemove_Password].m_sValue, nUID))
				{
					PGresult *xResult3 = NULL;
					// Remove the user from all the groups.
					if (PG_ExecA(xClient, &xResult3, g_xConfig[c_nQuery_UserRemove_Group].m_sValue, nUID))
					{
						PGresult *xResult4 = NULL;
						// Remove the user from the userlist.
						if (PG_ExecA(xClient, &xResult4, g_xConfig[c_nQuery_UserRemove_User].m_sValue, nUID))
						{
							fuzzd_Log(LOG_INFO, "[%s] User %s removed succesfully.\n", xClient->m_sHostname, xClient->m_xPkt_String.m_sString);
							
							// Call the cleanup script.
							bRet = true;
						}
						else fuzzd_Log(LOG_ERR, "[%s] User %s partially removed, couldn't remove user entry: %s.\n", xClient->m_sHostname, xClient->m_xPkt_String.m_sString, PQerrorMessage(g_xConn));
						if (xResult4) PQclear(xResult4);
					}
					else fuzzd_Log(LOG_ERR, "[%s] User %s partially removed, couldn't remove user from any groups: %s.\n", xClient->m_sHostname, xClient->m_xPkt_String.m_sString, PQerrorMessage(g_xConn));
					if (xResult3) PQclear(xResult3);
				}
				else fuzzd_Log(LOG_ERR, "[%s] User %s partially removed, couldn't remove user from password (shadow): %s.\n", xClient->m_sHostname, xClient->m_xPkt_String.m_sString, PQerrorMessage(g_xConn));
				if (xResult2) PQclear(xResult2);
			}
		}
		else fuzzd_Log(LOG_ERR, "[%s] Removing user %s failed: %s.\n", xClient->m_sHostname, xClient->m_xPkt_String.m_sString, PQerrorMessage(g_xConn));
		if (xResult) PQclear(xResult);
		free(sUsername);
	}
	else fuzzd_Log(LOG_ERR, "[%s] Couldn't allocate memory for escaping username\n", xClient->m_sHostname);
	return bRet;
}

/********************
 Change a User
********************/
bool pg_UserChange(struct Client *xClient)
{
	bool bRet = false;
	return bRet;
	xClient = xClient;
}

/********************
 Add a new Group
********************/
bool pg_GroupAdd(struct Client *xClient)
{
	bool bRet = false;
	return bRet;
	xClient = xClient;
}

/********************
 Remove a Group
********************/
bool pg_GroupRemove(struct Client *xClient)
{
	bool bRet = false;
	return bRet;
	xClient = xClient;
}

/********************
 Change a Group
********************/
bool pg_GroupChange(struct Client *xClient)
{
	bool bRet = false;
	return bRet;
	xClient = xClient;
}

/********************
 Add a new Member
********************/
bool pg_MemberAdd(struct Client *xClient)
{
	bool bRet = false;
	return bRet;
	xClient = xClient;
}

/********************
 Remove a Member
********************/
bool pg_MemberRemove(struct Client *xClient)
{
	bool bRet = false;
	return bRet;
	xClient = xClient;
}

/**************
 Authenticator
**************/
struct Authenticator postgresql_Authenticator = {
	pg_Initialize,
	pg_Shutdown,
	pg_GetInfo,
	pg_Authenticate,
	pg_ChangePassword,
	pg_GetUsers,
	pg_MatchUserID,
	pg_MatchUserName,
	pg_MatchUserGECOS,
	pg_GetGroups,	
	pg_MatchGroupID,
	pg_MatchGroupName,
	pg_UserAdd,	
	pg_UserRemove,
	pg_UserChange,
	pg_GroupAdd,
	pg_GroupRemove,
	pg_GroupChange,
	pg_MemberAdd,
	pg_MemberRemove
};
