/* FUZZd Relay Authenticator by Jeroen Massar <jeroen@unfix.org> */

// FUZZd
#include "fuzzd.h"
#include "fuzzauth_common.h"

// Globals
SOCKET	g_xRelayConn = SOCKET_ERROR;
int		g_nRelayConnectCount = 0;

/***********************
 Initialize the module
***********************/
bool relay_Initialize(void)
{
	int bRet = false;

	D(DLOG("relay_Initialize()\n"));
	if (fuzzauth_ConnectB((SOCKET *)&g_xRelayConn, (int *)&g_nRelayConnectCount, "heaven.unfix.org", FUZZAUTH_PORT, DEFAULT_FAMILY))
	{
		fuzzd_Log(LOG_INFO, "FUZZauth Relay connection succesful.\n");
		bRet = true;
		
		// Fetch & Parse first line.
		
		// Send MODE command and get into Packet/SSH or Packet/SSL mode.
	}
	else
	{
		fuzzd_Log(LOG_ERR, "FUZZauth Relay connection failed\n");
		bRet = false;
	}
	return bRet;
}

/*********************
 Shutdown the module
*********************/
bool relay_Shutdown(void)
{
	D(DLOG("relay_Shutdown()\n"));
	fuzzauth_DisconnectA((SOCKET *)&g_xRelayConn, (int *)&g_nRelayConnectCount);
	return true;
}

/******************************
 Get some information.
******************************/
bool relay_GetInfo(char *sBuf, int nSize)
{
	strncpy(sBuf, "FUZZauth Relay Authenticator Plugin by Jeroen Massar\n", nSize);
	return true;
}

/**************************
 Authenticate this Client
**************************/
bool relay_Authenticate(struct Client *xClient)
{
	bool			bRet = false;
	D(DLOG("[%s] relay_Authenticate()\n", xClient->m_sHostname));
	return bRet;
}

/****************
 Change Password
****************/
bool relay_ChangePassword(struct Client *xClient)
{
	bool bRet = false;
	D(DLOG("[%s] relay_ChangePassword()\n", xClient->m_sHostname));
	return bRet;
}

/****************************
 Get Users from our backend.
****************************/
bool relay_GetUsers(struct Client *xClient)
{
	bool		bRet = false;
	xClient->m_xPkt_Answer.m_nAnswer = c_nAnswer_NoMoreData;
	return bRet;
}

/*********************
 Match User by UserID
*********************/
bool relay_MatchUserID(struct Client *xClient)
{
	bool		bRet = false;
	if (!bRet) xClient->m_xPkt_Answer.m_nAnswer = c_nAnswer_Denied;
	return bRet;
}

/*******************
 Match User by name
*******************/
bool relay_MatchUserName(struct Client *xClient)
{
	bool		bRet = false;
	D(DLOG("[%s] relay_MatchUserName('%s')\n", xClient->m_sHostname, xClient->m_xPkt_String.m_sString));
	if (!bRet) xClient->m_xPkt_Answer.m_nAnswer = c_nAnswer_Denied;
	return bRet;
}

/*******************
 Match User by GECOS
*******************/
bool relay_MatchUserGECOS(struct Client *xClient)
{
	bool		bRet = false;
	D(DLOG("[%s] relay_MatchUserGECOS('%s')\n", xClient->m_sHostname, xClient->m_xPkt_String.m_sString));
	if (!bRet) xClient->m_xPkt_Answer.m_nAnswer = c_nAnswer_Denied;
	return bRet;
}

/*****************************
 Get Groups from our backend.
*****************************/
bool relay_GetGroups(struct Client *xClient)
{
	bool		bRet = false;
	D(DLOG("[%s] relay_GetUsers()\n", xClient->m_sHostname));
	xClient->m_xPkt_Answer.m_nAnswer = c_nAnswer_NoMoreData;
	return bRet;
}

/***********************
 Match Group by GroupID
***********************/
bool relay_MatchGroupID(struct Client *xClient)
{
	bool		bRet = false;
	D(DLOG("[%s] relay_MatchGroupID(%d)\n", xClient->m_sHostname, xClient->m_xPkt_Number.m_nNumber));
	if (!bRet) xClient->m_xPkt_Answer.m_nAnswer = c_nAnswer_Denied;
	return bRet;
}

/********************
 Match Group by name
********************/
bool relay_MatchGroupName(struct Client *xClient)
{
	bool		bRet = false;
	D(DLOG("[%s] relay_MatchGroupName('%s')\n", xClient->m_sHostname, xClient->m_xPkt_String.m_sString));
	if (!bRet) xClient->m_xPkt_Answer.m_nAnswer = c_nAnswer_Denied;
	return bRet;
}

/********************
 Add a new User
********************/
bool relay_UserAdd(struct Client *xClient)
{
	bool bRet = false;
	return bRet;
	xClient = xClient;
}

/********************
 Remove a User
********************/
bool relay_UserRemove(struct Client *xClient)
{
	bool bRet = false;
	return bRet;
	xClient = xClient;
}

/********************
 Change a User
********************/
bool relay_UserChange(struct Client *xClient)
{
	bool bRet = false;
	return bRet;
	xClient = xClient;
}

/********************
 Add a new Group
********************/
bool relay_GroupAdd(struct Client *xClient)
{
	bool bRet = false;
	return bRet;
	xClient = xClient;
}

/********************
 Remove a Group
********************/
bool relay_GroupRemove(struct Client *xClient)
{
	bool bRet = false;
	return bRet;
	xClient = xClient;
}

/********************
 Change a Group
********************/
bool relay_GroupChange(struct Client *xClient)
{
	bool bRet = false;
	return bRet;
	xClient = xClient;
}

/********************
 Add a new Member
********************/
bool relay_MemberAdd(struct Client *xClient)
{
	bool bRet = false;
	return bRet;
	xClient = xClient;
}

/********************
 Remove a Member
********************/
bool relay_MemberRemove(struct Client *xClient)
{
	bool bRet = false;
	return bRet;
	xClient = xClient;
}

/**************
 Authenticator
**************/
struct Authenticator relay_Authenticator = {
	relay_Initialize,
	relay_Shutdown,
	relay_GetInfo,
	relay_Authenticate,
	relay_ChangePassword,
	relay_GetUsers,
	relay_MatchUserID,
	relay_MatchUserName,
	relay_MatchUserGECOS,
	relay_GetGroups,	
	relay_MatchGroupID,
	relay_MatchGroupName,
	relay_UserAdd,	
	relay_UserRemove,
	relay_UserChange,
	relay_GroupAdd,
	relay_GroupRemove,
	relay_GroupChange,
	relay_MemberAdd,
	relay_MemberRemove
};
