/*	FUZZauth - http://unfix.org/projects/fuzzauth & http://fuzzauth.sourceforge.net
 * FUZZd - a FUZZauth Server daemon by Jeroen Massar <jeroen@unfix.org>
*/

#include "fuzzd.h"

/* Global variables */
volatile bool					g_bQuit = false, g_bSigHUP = false, g_bVerbose = false;
struct list						g_xClientList;
fd_set							g_xSockSet;
int								g_nHighSock, g_nNumSocks;
char								g_sServername[BUFFER_SIZE];
enum c_nEncryptionModes		g_nEncryption_Mode = c_nEncryption_MD5;
SOCKET							g_xServSock[FD_SETSIZE];
D(
int								g_nDebugLevel = 0;
)

/* Our backend */
struct Authenticator			*g_xAuthenticator = NULL;
void								*g_xBackend = NULL;

#define FUZZD_AUTH_MODULE "relay"

/************************
 Logging Function
************************/
void fuzzd_LogA(int nPriority, const char *sFormat, va_list xArgs)
{
D(
	if (g_nDebugLevel == 0)
	{
)
		openlog(FUZZD_NAME, LOG_PID, LOG_AUTH );
		vsyslog(nPriority, sFormat, xArgs);
		closelog();
D(
	}
	else
	{
		printf("[%-7s] ", nPriority == LOG_ERR ? "ERR" : (nPriority == LOG_INFO ? "INFO" : (nPriority == LOG_DEBUG ? "DEBUG" : (nPriority == LOG_WARNING ? "WARNING" : (nPriority == LOG_NOTICE ? "NOTICE" : "[???????]")))));
		vprintf(sFormat, xArgs);
	}
)
}

void fuzzd_Log(int nPriority, const char *sFormat, ... )
{
	va_list xArgs;
	va_start(xArgs, sFormat);
	fuzzd_LogA(nPriority, sFormat, xArgs);
	va_end(xArgs);
}

D(
/************************
 Debug Functions
************************/
void DLOG(const char *sFormat, ... )
{
	if (g_nDebugLevel > 0)
	{
		va_list xArgs;
		va_start(xArgs, sFormat);
		fuzzd_LogA(LOG_DEBUG, sFormat, xArgs);
		va_end(xArgs);
	}
}
)

/************************
 Socket printf
************************/
bool soprintf(struct Client *xClient, char *sFormat, ...) __attribute__((format(printf,2,3)));
bool soprintf(struct Client *xClient, char *sFormat, ...)
{
	bool bRet = true;
	if (xClient->m_xSocket != INVALID_SOCKET)
	{
		va_list	xArg;
		char		*sBuf;
		int		nLen;
		va_start(xArg, sFormat);
		nLen = vasprintf(&sBuf, sFormat, xArg);
		va_end(xArg);
		if (sBuf)
		{
			if (write(xClient->m_xSocket, sBuf, nLen) == SOCKET_ERROR) bRet = false;
			free(sBuf);
		}
	}
	else bRet = false;
	return bRet;
}

/************************
 Signal Handler
************************/
void sigint(int i);
void signal_init(void)
{
#ifdef SIGALRM
	signal(SIGALRM, &sigint);
#endif
#ifdef SIGHUP
	signal(SIGHUP, &sigint);
#endif
#ifdef SIGINT
	signal(SIGINT, &sigint);
#endif
#ifdef SIGKILL
	signal(SIGKILL, &sigint);
#endif
#ifdef SIGTERM
	signal(SIGTERM, &sigint);
#endif
#ifdef SIGABRT
	signal(SIGABRT, &sigint);
#endif
#ifdef SIGFPE
	signal(SIGFPE, &sigint);
#endif
#ifdef SIGILL
	signal(SIGILL, &sigint);
#endif
#ifdef SIGQUIT
	signal(SIGQUIT, &sigint);
#endif
#ifdef SIGSEGV
	signal(SIGSEGV, &sigint);
#endif
#ifdef SIGTRAP
	signal(SIGTRAP, &sigint);
#endif
}

void sigint(int i)
{
	// Check if this was a SigHUP, which will restart us completely.
	if (i == SIGHUP) g_bSigHUP = true;

	g_bQuit = true;
	/* Re-set it at all once, to catch following interrupts too.*/
	signal_init();

	fuzzd_Log(LOG_INFO, "Received interrupt %d... %s...\n", i, i == SIGHUP ? "reloading" : "quitting");
}

/***************************
 Set a socket non blocking.
***************************/
void net_set_non_blocking(int sock)
{
	struct timeval maxtime;
 	struct linger	ling;				/* for linger */

	/* Unix. */
#if !defined(WIN32)
	int res,nonb=0;
#endif
#if !defined (WIN32) && defined(O_NONBLOCK)
	nonb|=O_NONBLOCK;
#endif
#if !defined (WIN32) && defined(O_NDELAY)
	nonb|=O_NDELAY;
#endif
#if !defined (WIN32) && defined(NBLOCK_SYSV)
	res = 1;
	if (ioctl(sock,FIONBIO,&res)<0) { } fuzzd_Log(LOG_ERR, "ioctl(sock,FIONBIO) failed...\n");
#endif
#if !defined (WIN32) && !defined(NBLOCK_SYSV)
	if ((res=fcntl(sock,F_GETFL,0))==-1) fuzzd_Log(LOG_ERR, "fcntl(sock, F_GETFL) failed...\n");
	else if (fcntl(sock,F_SETFL,res|nonb)==-1) fuzzd_Log(LOG_ERR, "fcntl(sock, F_SETL, nonb) failed...\n");
#endif

	/* Windows. */
#if defined(WIN32)
	unsigned long ret=0;
#endif
#if defined(WIN32) && defined(FIONBIO)
	ret=1;
	ioctlsocket(sock,FIONBIO,&ret);
#endif
#if defined(WIN32) && !defined(FIONBIO)
	if ((ret=fcntl(sock,F_GETFL,0))!=-1) fcntl(sock,F_SETFL,ret|nonb);
#endif

	/* Unix & Windows. */
	maxtime.tv_sec=0;
	maxtime.tv_usec=10;
	setsockopt(sock,SOL_SOCKET,SO_RCVTIMEO,(const char *)&maxtime,sizeof(maxtime));
	maxtime.tv_sec=0;
	maxtime.tv_usec=10;
	setsockopt(sock,SOL_SOCKET,SO_SNDTIMEO,(const char *)&maxtime,sizeof(maxtime));

 	/* Use SO_LINGER so it won't hang up on client disconnects -> Closed == CLOSED :) */
 	ling.l_onoff=1;					/* Turn it on. */
 	ling.l_linger=0;					/* But NO timeout. */
 	setsockopt(sock,SOL_SOCKET,SO_LINGER,(const char *)&ling,sizeof(struct linger));
}

/************************
 Accept a client
************************/
void AcceptClient(SOCKET xServer, bool bUnix)
{
	unsigned int	nFromLen;
	struct Client	*xClient = NULL;

	xClient = malloc(sizeof(struct Client));
	if (xClient)
	{
		/* Setup Client */
		memset(xClient,0, sizeof(struct Client));
		xClient->m_bUnix = bUnix;
		xClient->m_nMode = bUnix ? c_nMode_Packet : c_nMode_Human;
		xClient->m_bGotCommand = false;

		nFromLen = sizeof(xClient->m_xFrom);
		xClient->m_xSocket = accept(xServer, (struct sockaddr *)&xClient->m_xFrom, &nFromLen);
		if (xClient->m_xSocket != INVALID_SOCKET)
		{
			char *sHostname = xClient->m_sHostname;
			int nLen = sizeof(xClient->m_sHostname)-1;
			sHostname[0] = '\0';
			if (getnameinfo((struct sockaddr *)&xClient->m_xFrom, nFromLen, sHostname, nLen, NULL, 0, NI_NAMEREQD) == 0)
			{
				nLen			-=		strlen(sHostname);
				sHostname	+=		strlen(sHostname);
				*sHostname	++=	' ';
				*sHostname	++=	'(';
			}
			if (getnameinfo((struct sockaddr *)&xClient->m_xFrom, nFromLen, sHostname,	nLen,	NULL, 0, NI_NUMERICHOST) != 0) strncpy(sHostname, "<unknown>", nLen);
			if (sHostname > xClient->m_sHostname)
			{
				sHostname	+=		strlen(sHostname);
				*sHostname	++=	')';
				*sHostname	++=	'\0';
			}
			// Also get the peercredentials for this user when it's a unixsocket.
			if (bUnix)
			{
				struct ucred cr;
				unsigned int cl = sizeof(cr);
				sHostname--;
				if (getsockopt(xClient->m_xSocket, SOL_SOCKET, SO_PEERCRED, &cr, &cl) == 0)
				{
					snprintf(sHostname, nLen, " pid=%d, uid=%d, gid=%d", cr.pid, cr.uid, cr.gid);
				}
			}

			/* Add the Client to our ClientList */
			lst_AddTail(&g_xClientList, xClient);
			/* And set it up to be serviced... */
			if (xClient->m_xSocket > g_nHighSock) g_nHighSock = xClient->m_xSocket;
			FD_SET(xClient->m_xSocket, &g_xSockSet);
			net_set_non_blocking(xClient->m_xSocket);

			D(DLOG("[%s] =======================================================\n", xClient->m_sHostname));
			if ((!bUnix) || ((bUnix) && (g_bVerbose)) D(|| (g_nDebugLevel > 0) ) ) fuzzd_Log(LOG_INFO, "[%s] Accepted connection\n", xClient->m_sHostname);

			// Only in human mode we say hello.
			if (xClient->m_nMode == c_nMode_Human) soprintf(xClient, "200 %s %s v%d\r\n", g_sServername, FUZZAUTH_NAME, FUZZAUTH_VERSION);
		}
		else
		{
			fuzzd_Log(LOG_ERR, "accept() failed with error %d: %s\n", ERRORNUMSTR);
			free(xClient);
			xClient = NULL;
		}
	}
}

// Internal Command for checking the adminname+password supplied in the command packet.
bool Cmd_AdminCheck(struct Client *xClient)
{
	bool bRet = true;
	fuzzd_Log(LOG_INFO, "[%s] Is executing an administrative command\n", xClient->m_sHostname);
	return bRet;
}

/***********************************************
 Backend Connection
***********************************************/
bool Cmd_Authenticate		(struct Client *xClient)	{	return 																			g_xAuthenticator->cmd_Authenticate		(xClient);	}
bool Cmd_GetUsers				(struct Client *xClient)	{	return 																			g_xAuthenticator->cmd_GetUsers			(xClient);	}
bool Cmd_MatchUserID			(struct Client *xClient)	{	return 																			g_xAuthenticator->cmd_MatchUserID		(xClient);	}
bool Cmd_MatchUserName		(struct Client *xClient)	{	return 																			g_xAuthenticator->cmd_MatchUserName		(xClient);	}
bool Cmd_MatchUserGECOS		(struct Client *xClient)	{	return 																			g_xAuthenticator->cmd_MatchUserGECOS	(xClient);	}
bool Cmd_GetGroups			(struct Client *xClient)	{	return 																			g_xAuthenticator->cmd_GetGroups			(xClient);	}
bool Cmd_MatchGroupID		(struct Client *xClient)	{	return 																			g_xAuthenticator->cmd_MatchGroupID		(xClient);	}
bool Cmd_MatchGroupName		(struct Client *xClient)	{	return 																			g_xAuthenticator->cmd_MatchGroupName	(xClient);	}
bool Cmd_ChangePassword		(struct Client *xClient)	{	return (Cmd_Authenticate(xClient)||Cmd_AdminCheck(xClient))	&&	g_xAuthenticator->cmd_ChangePassword	(xClient);	}
bool Cmd_UserAdd				(struct Client *xClient)	{	return Cmd_AdminCheck(xClient)										&&	g_xAuthenticator->cmd_UserAdd				(xClient);	}
bool Cmd_UserRemove			(struct Client *xClient)	{	return Cmd_AdminCheck(xClient)										&&	g_xAuthenticator->cmd_UserRemove			(xClient);	}
bool Cmd_UserChange			(struct Client *xClient)	{	return Cmd_AdminCheck(xClient)										&&	g_xAuthenticator->cmd_UserChange			(xClient);	}
bool Cmd_GroupAdd				(struct Client *xClient)	{	return Cmd_AdminCheck(xClient)										&&	g_xAuthenticator->cmd_GroupAdd			(xClient);	}
bool Cmd_GroupRemove			(struct Client *xClient)	{	return Cmd_AdminCheck(xClient)										&&	g_xAuthenticator->cmd_GroupRemove		(xClient);	}
bool Cmd_GroupChange			(struct Client *xClient)	{	return Cmd_AdminCheck(xClient)										&&	g_xAuthenticator->cmd_GroupChange		(xClient);	}
bool Cmd_MemberAdd			(struct Client *xClient)	{	return Cmd_AdminCheck(xClient)										&&	g_xAuthenticator->cmd_MemberAdd			(xClient);	}
bool Cmd_MemberRemove		(struct Client *xClient)	{	return Cmd_AdminCheck(xClient)										&&	g_xAuthenticator->cmd_MemberRemove		(xClient);	}

/*******************************
 Password Encryption Functions
*******************************/
// Generate an radix-64 character from an integer.
int FD_i64c(int i)
{
	return	i <= 0					?	'.'				:
				i == 1					?	'/'				:
				i >= 2	&& i < 12	?	'0' - 2 + i		:
				i >= 12	&& i < 38	?	'A' - 12 + i	:
				i >= 38	&& i < 63	?	'a' - 38 + i	:
												'z';
}

/* Generate random salt character. */
char *FD_GenerateSalt(void)
{
	time_t					xNow;
	static unsigned long	x;
	static char				sResult[3];

	time(&xNow);
	x += xNow + getpid() + clock();
	sResult[0] = FD_i64c(((x >> 18) ^ (x >> 6)) & 077);
	sResult[1] = FD_i64c(((x >> 12) ^ x) & 077);
	sResult[2] = '\0';
	return sResult;
}

// The encrypted value should be free()'d
// This should be used for new passwords.
char *fuzzd_Encrypt(char *sPass)
{
	char 	*s = NULL;
#ifdef FA_CRYPT_MHASH
	int	nMHashType = -1;
#endif

	switch (g_nEncryption_Mode)
	{
		case c_nEncryption_Crypt:		s = strdup((char *)crypt(sPass, FD_GenerateSalt()));
												break;
#ifdef FA_CRYPT_MHASH
		case c_nEncryption_MD5:			nMHashType = MHASH_MD5;			break;
		case c_nEncryption_CRC32:		nMHashType = MHASH_CRC32;		break;
		case c_nEncryption_CRC32B:		nMHashType = MHASH_CRC32B;		break;
		case c_nEncryption_SHA1:		nMHashType = MHASH_SHA1;		break;
		case c_nEncryption_SHA256:		nMHashType = MHASH_SHA256;		break;
		case c_nEncryption_HAVAL256:	nMHashType = MHASH_HAVAL256;	break;
		case c_nEncryption_HAVAL192:	nMHashType = MHASH_HAVAL192;	break;
		case c_nEncryption_HAVAL160:	nMHashType = MHASH_HAVAL160;	break;
		case c_nEncryption_HAVAL128:	nMHashType = MHASH_HAVAL128;	break;
		case c_nEncryption_RIPEMD160:	nMHashType = MHASH_RIPEMD160;	break;
		case c_nEncryption_TIGER:		nMHashType = MHASH_TIGER;		break;
		case c_nEncryption_TIGER160:	nMHashType = MHASH_TIGER160;	break;
		case c_nEncryption_TIGER128:	nMHashType = MHASH_TIGER128;	break;
		case c_nEncryption_GOST:		nMHashType = MHASH_GOST;		break;
												break;
#endif
		case c_nEncryption_Clear:
												s = strdup(sPass);
	}
#ifdef FA_CRYPT_MHASH
	if (nMHashType != -1)
	{
		char				*sBuf;
		int				nBufSize;
		MHASH				xHandle;
		unsigned char	*sHash;
		
		xHandle = mhash_init(nMHashType);
		if (xHandle == MHASH_FAILED)
		{
			fuzzd_Log(LOG_ERR, "fuzzd_Encrypt() - Couldn't initialize mhash library!\n");
		}
		else
		{
			unsigned int i;
			// sPass comes from our own functions which have checked it already ofcourse ;)
			mhash(xHandle, sPass, strlen(sPass));
			sHash = mhash_end(xHandle);
			nBufSize = (mhash_get_block_size(nMHashType) * 2) + 1 + 3; // 1='\0', 3='$n$';
			sBuf = (char *)malloc(nBufSize);
			memset(sBuf, 0, nBufSize);
			snprintf(sBuf, 4, "$%d$", nMHashType);
			for (i = 0; i < mhash_get_block_size(nMHashType); i++)
			{
				snprintf(&sBuf[i * 2] + 3, nBufSize - (i*2) - 3, "%.2x", sHash[i]);
			}
			s = sBuf;
		}
	}
#endif
	return s;
}

bool fuzzd_ComparePassword(char *sStored, char *sInput)
{
	bool	bRet = false;
#ifdef FA_CRYPT_MHASH
	int	nMHashType = -1;
#endif

	switch (g_nEncryption_Mode)
	{											// sStored comes from a database, so it's checked.
		case c_nEncryption_Crypt:		if (strlen(sStored) != 13)
												{
													fuzzd_Log(LOG_ERR, "fuzzd_ComparePassword() - Stored password has invalid length.\n");
												}
												else
												{
													char sSalt[3];
													char *sCrypt = crypt(sInput, sSalt);
													sSalt[0] = sStored[0];
													sSalt[1] = sStored[1];
													sSalt[2] = '\0';
													bRet = (strcmp(sStored, sCrypt) == 0) ? true : false;
													free(sCrypt);
												}
												break;
#ifdef FA_CRYPT_MHASH
		case c_nEncryption_MD5:			nMHashType = MHASH_MD5;			break;
		case c_nEncryption_CRC32:		nMHashType = MHASH_CRC32;		break;
		case c_nEncryption_CRC32B:		nMHashType = MHASH_CRC32B;		break;
		case c_nEncryption_SHA1:		nMHashType = MHASH_SHA1;		break;
		case c_nEncryption_SHA256:		nMHashType = MHASH_SHA256;		break;
		case c_nEncryption_HAVAL256:	nMHashType = MHASH_HAVAL256;	break;
		case c_nEncryption_HAVAL192:	nMHashType = MHASH_HAVAL192;	break;
		case c_nEncryption_HAVAL160:	nMHashType = MHASH_HAVAL160;	break;
		case c_nEncryption_HAVAL128:	nMHashType = MHASH_HAVAL128;	break;
		case c_nEncryption_RIPEMD160:	nMHashType = MHASH_RIPEMD160;	break;
		case c_nEncryption_TIGER:		nMHashType = MHASH_TIGER;		break;
		case c_nEncryption_TIGER160:	nMHashType = MHASH_TIGER160;	break;
		case c_nEncryption_TIGER128:	nMHashType = MHASH_TIGER128;	break;
		case c_nEncryption_GOST:		nMHashType = MHASH_GOST;		break;
												break;
#endif
		case c_nEncryption_Clear:
												bRet = strcmp(sStored, sInput) == 0 ? true : false;
	}

#ifdef FA_CRYPT_MHASH
	if (nMHashType != -1)
	{
		char				*sCrypt;
		int				nBufSize;
		MHASH				xHandle;
		unsigned char	*sHash;
		
		xHandle = mhash_init(nMHashType);
		if (xHandle == MHASH_FAILED)
		{
			fuzzd_Log(LOG_ERR, "fuzzd_ComparePassword() - Couldn't initialize mhash library!\n");
		}
		else
		{
			unsigned int i;
			// sInput comes from our own functions.
			mhash(xHandle, sInput, strlen(sInput));
			sHash = mhash_end(xHandle);
			nBufSize = (mhash_get_block_size(nMHashType) * 2) + 1 + 3; // 1='\0', 3='$n$';
			sCrypt = (char *)malloc(nBufSize);
			memset(sCrypt, 0, nBufSize);
			snprintf(sCrypt, 4, "$%d$", nMHashType);
			for (i = 0; i < mhash_get_block_size(nMHashType); i++)
			{
				snprintf(&sCrypt[i * 2] + 3, nBufSize - (i*2) - 3, "%.2x", sHash[i]);
			}
			D(DLOG("fuzzd_ComparePassword() - MHASH(%d) - Comparing '%s' with '%s'\n", nMHashType, sStored, sCrypt));
			bRet = (strcmp(sStored, sCrypt) == 0) ? true : false;
			free(sCrypt);
		}
	}
#endif
	return bRet;
}

/*******************************
 Return an Answer to the client
*******************************/
bool fuzzd_ReturnAnswer(struct Client *xClient)
{
	bool	bRet = false;
	void	*xPkt = NULL;
D(
	char	*sName = NULL;

	switch (xClient->m_xPkt_Answer.m_nAnswer)
	{
		case c_nAnswer_Unknown:				sName = c_sAnswer_Unknown;				break;
		case c_nAnswer_UnknownVersion:	sName = c_sAnswer_UnknownVersion;	break;
		case c_nAnswer_UnknownMode:		sName = c_sAnswer_UnknownMode;		break;
		case c_nAnswer_UnknownCommand:	sName = c_sAnswer_UnknownCommand;	break;
		case c_nAnswer_NeedDataPacket:	sName = c_sAnswer_NeedDataPacket;	break;
		case c_nAnswer_NoMoreData:			sName = c_sAnswer_NoMoreData;			break;
		case c_nAnswer_AlreadySent:		sName = c_sAnswer_AlreadySent;		break;
		case c_nAnswer_Granted:				sName = c_sAnswer_Granted;				break;
		case c_nAnswer_Denied:				sName = c_sAnswer_Denied;				break;
	}
)
	if ((xClient->m_xPkt_Answer.m_nAnswer != c_nAnswer_AlreadySent) && (xClient->m_xPkt_Answer.m_nAnswer != c_nAnswer_Unknown))
	{
		// Only "Granted" will pass back data.
		if (xClient->m_xPkt_Answer.m_nAnswer != c_nAnswer_Granted) xClient->m_xPkt_Answer.m_nDatatype = c_nDatatype_Unknown;
	
		switch (xClient->m_xPkt_Answer.m_nDatatype)
		{
			case c_nDatatype_Number:	xPkt = &xClient->m_xPkt_Number;
												xClient->m_xPkt_Answer.m_nLength = sizeof(struct faPacket_Number);
												D(DLOG("[%s] ReturnAnswer() - Number: %d\n", xClient->m_sHostname, xClient->m_xPkt_Number.m_nNumber));
												xClient->m_xPkt_Number.m_nNumber = htonl(xClient->m_xPkt_Number.m_nNumber);
												break;
			case c_nDatatype_Number2:	xPkt = &xClient->m_xPkt_Number2;
												xClient->m_xPkt_Answer.m_nLength = sizeof(struct faPacket_Number2);
												D(DLOG("[%s] ReturnAnswer() - Number2: %d, %d\n", xClient->m_sHostname, xClient->m_xPkt_Number2.m_nNumber1, xClient->m_xPkt_Number2.m_nNumber2));
												xClient->m_xPkt_Number2.m_nNumber1 = htonl(xClient->m_xPkt_Number2.m_nNumber1);
												xClient->m_xPkt_Number2.m_nNumber2 = htonl(xClient->m_xPkt_Number2.m_nNumber2);
												break;
			case c_nDatatype_String:	xPkt = &xClient->m_xPkt_String;
												xClient->m_xPkt_Answer.m_nLength = sizeof(struct faPacket_String);
												D(DLOG("[%s] ReturnAnswer() - String: %s\n", xClient->m_sHostname, xClient->m_xPkt_String.m_sString));
												break;
			case c_nDatatype_User:		xPkt = &xClient->m_xPkt_User;
												xClient->m_xPkt_Answer.m_nLength = sizeof(struct faPacket_User);
												D(DLOG("[%s] ReturnAnswer() - User: %d:%s:%s\n", xClient->m_sHostname, xClient->m_xPkt_User.m_nUID, xClient->m_xPkt_User.m_sUsername, xClient->m_xPkt_User.m_sGECOS));
												xClient->m_xPkt_User.m_nUID = htonl(xClient->m_xPkt_User.m_nUID);
												xClient->m_xPkt_User.m_nGID = htonl(xClient->m_xPkt_User.m_nGID);
												break;
			case c_nDatatype_Group:		xPkt = &xClient->m_xPkt_Group;
												xClient->m_xPkt_Answer.m_nLength = sizeof(struct faPacket_Group);
												D(DLOG("[%s] ReturnAnswer() - Group: %d:%s:%d\n", xClient->m_sHostname, xClient->m_xPkt_Group.m_nGID, xClient->m_xPkt_Group.m_sGroupname, xClient->m_xPkt_Group.m_nNumMembers));
												xClient->m_xPkt_Group.m_nGID = htonl(xClient->m_xPkt_Group.m_nGID);
												xClient->m_xPkt_Group.m_nNumMembers = htonl(xClient->m_xPkt_Group.m_nNumMembers);
												break;
			case c_nDatatype_Unknown:	xPkt = NULL;
												xClient->m_xPkt_Answer.m_nLength = 0;
												break;
		}
		D(DLOG("[%s] ReturnAnswer() - Returning Answer %s (Data=%x, Len=%d)\n", xClient->m_sHostname, sName, (unsigned int)xPkt, xClient->m_xPkt_Answer.m_nLength));

		// Fixup endianness in the Answer packet.
		xClient->m_xPkt_Answer.m_nAnswer		= htonl(xClient->m_xPkt_Answer.m_nAnswer);
		xClient->m_xPkt_Answer.m_nDatatype	= htonl(xClient->m_xPkt_Answer.m_nDatatype);
		xClient->m_xPkt_Answer.m_nLength		= htonl(xClient->m_xPkt_Answer.m_nLength);

		if (write(xClient->m_xSocket, &xClient->m_xPkt_Answer, sizeof(xClient->m_xPkt_Answer)) != SOCKET_ERROR)
		{
			if (xClient->m_xPkt_Answer.m_nLength != 0)
			{
				D(DLOG("[%s] ReturnAnswer() - Returning Answer (Data)\n", xClient->m_sHostname));
				if (write(xClient->m_xSocket, xPkt, ntohl(xClient->m_xPkt_Answer.m_nLength)) != SOCKET_ERROR)
				{
					bRet = true;
					D(DLOG("[%s] ReturnAnswer() - Data sent succesfully: %d: %s\n", xClient->m_sHostname, ERRORNUMSTR));
				}
				else fuzzd_Log(LOG_ERR, "[%s] ReturnAnswer() - Socket error while trying to send answer data: %d: %s\n", xClient->m_sHostname, ERRORNUMSTR);
			}
			else bRet = true;
		}
		else fuzzd_Log(LOG_ERR, "[%s] ReturnAnswer() - Socket error while trying to send answer: %d: %s\n", xClient->m_sHostname, ERRORNUMSTR);
	}
	else
	{
		D(DLOG("[%s] ReturnAnswer() - Answer is either unknown or alreadysent (%d)\n", xClient->m_sHostname, xClient->m_xPkt_Answer.m_nAnswer));
		bRet = true;
	}
	return bRet;
}

/**********************************
 Handle Client Plaintext commands
**********************************/
bool Client_Plaintext_Command(struct Client *xClient)
{
	bool	bStayAlive = true;
	char	*sPtr, *sCmd, *sOpt;

	sCmd = strtok_r(xClient->m_sBuffer, " ", &sPtr);
	if (sCmd != NULL)
	{
		sOpt = strtok_r(NULL, " ", &sPtr);
	
		if (strcasecmp(sCmd, "QUIT") == 0)
		{
			bStayAlive = soprintf(xClient, "500 See you later aligator\r\n");
			bStayAlive = false;
		}
		else if (strcasecmp(sCmd, "HELP") == 0)
		{
			bStayAlive = soprintf(xClient, "204 ABOUT AUTHENTICATE MODE PASSWORD USERNAME VERSION QUIT\r\n");
		}
		else if (strcasecmp(sCmd, "MODE") == 0)
		{
			if ((sOpt != NULL) && (strlen(sOpt) > 0))
			{
				int nMode = c_nMode_Unknown;
				if			(strcasecmp(sOpt, c_sMode_Human		) == 0)	nMode = c_nMode_Human;
				else if	(strcasecmp(sOpt, c_sMode_Packet		) == 0)	nMode = c_nMode_Packet;
//				else if	(strcasecmp(sOpt, c_sMode_Packet_SSH) == 0)	nMode = c_nMode_Packet_SSH;
//				else if	(strcasecmp(sOpt, c_sMode_Packet_SSL) == 0)	nMode = c_nMode_Packet_SSL;
				
				if (nMode != c_nMode_Unknown && nMode != c_nMode_Packet_SSH && nMode != c_nMode_Packet_SSL)
				{
					bStayAlive = soprintf(xClient, "200 Mode entered\r\n");
					xClient->m_nMode = nMode;
				}
				else bStayAlive = soprintf(xClient, "300 Mode '%s' not supported\r\n", sOpt);
			}
			else bStayAlive = soprintf(xClient, "204 %s %s\r\n", c_sMode_Human, c_sMode_Packet); //, c_sMode_Packet_SSH, c_sMode_Packet_SSL);
		}
		else if (strcasecmp(sCmd, "VERSION") == 0)
		{
			bStayAlive = soprintf(xClient, "200 %d\r\n", FUZZAUTH_VERSION);
		}
		else if (strcasecmp(sCmd, "ABOUT") == 0)
		{
			bStayAlive = soprintf(xClient, "201 ABOUT\r\n"\
									"202 %s v%d server\r\n"\
									"202 Providing a PAM/NSS relay/proxy service for easy and secure distributed authentication\r\n"\
									"202 Website: %s\r\n"\
									"203 ABOUT\r\n", FUZZAUTH_NAME, FUZZAUTH_VERSION, FUZZAUTH_WEBSITE);
		}
		else if (strcasecmp(sCmd, "USERNAME") == 0)
		{
			if ((sOpt != NULL) && (strlen(sOpt) > 0))
			{
				strncpy(xClient->m_xPkt_User.m_sUsername, sOpt, sizeof(xClient->m_xPkt_User.m_sUsername)-1);
				bStayAlive = soprintf(xClient, "200 And your password please...\r\n");
			}
			else bStayAlive = soprintf(xClient, "300 And who might you be?\r\n");
		}
		else if (strcasecmp(sCmd, "PASSWORD") == 0)
		{
			if ((sOpt != NULL) && (strlen(sOpt) > 0))
			{
				strncpy(xClient->m_xPkt_User.m_sPassword, sOpt, sizeof(xClient->m_xPkt_User.m_sPassword)-1);
				bStayAlive = soprintf(xClient, "200 You may now try to authenticate\r\n");
			}
			else bStayAlive = soprintf(xClient, "300 I really don't think so...\r\n");
		}
		else if (strcasecmp(sCmd, "AUTHENTICATE") == 0)
		{
			// m_sUsername & m_sPassword are filled in above...
			if (strlen(xClient->m_xPkt_User.m_sUsername) > 0)
			{
				if (strlen(xClient->m_xPkt_User.m_sPassword) > 0)
				{
					xClient->m_xPkt_Request.m_nVersion	= FUZZAUTH_VERSION;
					xClient->m_xPkt_Request.m_nCommand = c_nCommand_Authenticate;
					xClient->m_xPkt_Request.m_nLength = sizeof(struct faPacket_User);
					if (Cmd_Authenticate(xClient))
					{
						bStayAlive = soprintf(xClient, "200 Username/Password combination match - authentication succesful\r\n");
					}
					bStayAlive = soprintf(xClient, "300 Username/Password combination MISmatch - authentication failed\r\n");
				}
				else bStayAlive = soprintf(xClient, "300 Would you mind sharing your password?\r\n");
			}
			else bStayAlive = soprintf(xClient, "300 Would you mind telling me who you are first?\r\n");
		}
		else bStayAlive = soprintf(xClient, "300 I don't know what you mean with '%s'\r\n", xClient->m_sBuffer);
	}
	else
	{
		bStayAlive = false;
		fuzzd_Log(LOG_ERR, "[%s] Disconnecting client due to abuse...\n", xClient->m_sHostname);
		soprintf(xClient, "500 Je ne pas comprende ca.\r\n");
	}
	return bStayAlive;
}

bool Validate_Packet_Number(struct Client *xClient, unsigned int *nAmountRead)
{
	bool bRet = false;
	if ((*nAmountRead) >= sizeof(struct faPacket_Number))
	{
		xClient->m_bGotCommand = false;
		memcpy(&xClient->m_xPkt_Number, &xClient->m_sBuffer, sizeof(struct faPacket_Number));
		*nAmountRead -= sizeof(struct faPacket_Number);
		memmove(xClient->m_sBuffer, &xClient->m_sBuffer[sizeof(struct faPacket_Number)], *nAmountRead);
		// There isn't much to validate about this.
		// Fix endianness.
		xClient->m_xPkt_Number.m_nNumber = ntohl(xClient->m_xPkt_Number.m_nNumber);
		bRet = true;
	}
	D(else DLOG("[%s] Validate_Packet_Number() - Datalength isn't sufficient for a number.\n", xClient->m_sHostname));
	return bRet;
}

bool Validate_Packet_Number2(struct Client *xClient, unsigned int *nAmountRead)
{
	bool bRet = false;
	if ((*nAmountRead) >= sizeof(struct faPacket_Number2))
	{
		xClient->m_bGotCommand = false;
		memcpy(&xClient->m_xPkt_Number2, &xClient->m_sBuffer, sizeof(struct faPacket_Number2));
		*nAmountRead -= sizeof(struct faPacket_Number2);
		memmove(xClient->m_sBuffer, &xClient->m_sBuffer[sizeof(struct faPacket_Number2)], *nAmountRead);
		// There isn't much to validate about this.
		// Fix endianness.
		xClient->m_xPkt_Number2.m_nNumber1 = ntohl(xClient->m_xPkt_Number2.m_nNumber1);
		xClient->m_xPkt_Number2.m_nNumber2 = ntohl(xClient->m_xPkt_Number2.m_nNumber2);
		bRet = true;
	}
	D(else DLOG("[%s] Validate_Packet_Number2() - Datalength isn't sufficient for a number2.\n", xClient->m_sHostname));
	return bRet;
}

bool Validate_Packet_String(struct Client *xClient, unsigned int *nAmountRead, bool *bStayAlive)
{
	bool bRet = false;
	if ((*nAmountRead) >= sizeof(struct faPacket_String))
	{
		unsigned int i;
		xClient->m_bGotCommand = false;
		memcpy(&xClient->m_xPkt_String, &xClient->m_sBuffer, sizeof(struct faPacket_String));
		*nAmountRead -= sizeof(struct faPacket_String);
		memmove(xClient->m_sBuffer, &xClient->m_sBuffer[sizeof(struct faPacket_String)], *nAmountRead);
		// Validate the packet contents.
		i = 0;
		while ((i < (sizeof(xClient->m_xPkt_String.m_sString)-1)) && (xClient->m_xPkt_String.m_sString[i] !='\0')) i++;
		// We deny if the string isn't NUL ('\0') terminated in our boundaries.
		if (i < (sizeof(xClient->m_xPkt_String.m_sString)-1))
		{
			bRet = true;
		}
		else
		{
			D(DLOG("[%s] Validate_Packet_String() - Sent us a non-NUL terminated string.\n", xClient->m_sHostname));
			*bStayAlive = false;
		}
	}
	D(else DLOG("[%s] Validate_Packet_String() - Datalength isn't sufficient for a string.\n", xClient->m_sHostname));
	return bRet;
}

bool Validate_Packet_User(struct Client *xClient, unsigned int *nAmountRead, bool *bStayAlive)
{
	bool bRet = false;
	if ((*nAmountRead) >= sizeof(struct faPacket_User))
	{
		unsigned int i;
		xClient->m_bGotCommand = false;
		memcpy(&xClient->m_xPkt_User, &xClient->m_sBuffer, sizeof(struct faPacket_User));
		*nAmountRead -= sizeof(struct faPacket_User);
		memmove(xClient->m_sBuffer, &xClient->m_sBuffer[sizeof(struct faPacket_User)], *nAmountRead);
		// Validate the packet contents.
		i = 0;
		while ((i < (sizeof(xClient->m_xPkt_User.m_sUsername)-1)) && (xClient->m_xPkt_User.m_sUsername[i] !='\0')) i++;
		if (i < (sizeof(xClient->m_xPkt_User.m_sUsername)-1))
		{
			i = 0;
			while ((i < (sizeof(xClient->m_xPkt_User.m_sPassword)-1)) && (xClient->m_xPkt_User.m_sPassword[i] !='\0')) i++;
			if (i < (sizeof(xClient->m_xPkt_User.m_sPassword)-1))
			{
				while ((i < (sizeof(xClient->m_xPkt_User.m_sPasswordNew)-1)) && (xClient->m_xPkt_User.m_sPasswordNew[i] !='\0')) i++;
				if (i < (sizeof(xClient->m_xPkt_User.m_sPasswordNew)-1))
				{
					while ((i < (sizeof(xClient->m_xPkt_User.m_sGECOS)-1)) && (xClient->m_xPkt_User.m_sGECOS[i] !='\0')) i++;
					if (i < (sizeof(xClient->m_xPkt_User.m_sGECOS)-1))
					{
						while ((i < (sizeof(xClient->m_xPkt_User.m_sHomedir)-1)) && (xClient->m_xPkt_User.m_sHomedir[i] !='\0')) i++;
						if (i < (sizeof(xClient->m_xPkt_User.m_sHomedir)-1))
						{
							while ((i < (sizeof(xClient->m_xPkt_User.m_sShell)-1)) && (xClient->m_xPkt_User.m_sShell[i] !='\0')) i++;
							if (i < (sizeof(xClient->m_xPkt_User.m_sShell)-1))
							{
								// Fix endianness.
								xClient->m_xPkt_User.m_nUID = ntohl(xClient->m_xPkt_User.m_nUID);
								xClient->m_xPkt_User.m_nGID = ntohl(xClient->m_xPkt_User.m_nGID);
								bRet = true;
							}
							else
							{
								D(DLOG("[%s] Validate_Packet_User() - Sent us a non-NUL terminated Shell.\n", xClient->m_sHostname));
								*bStayAlive = false;
							}
						}
						else
						{
							D(DLOG("[%s] Validate_Packet_User() - Sent us a non-NUL terminated Homedir.\n", xClient->m_sHostname));
							*bStayAlive = false;
						}
					}
					else
					{
						D(DLOG("[%s] Validate_Packet_User() - Sent us a non-NUL terminated GECOS.\n", xClient->m_sHostname));
						*bStayAlive = false;
					}
				}
				else
				{
					D(DLOG("[%s] Validate_Packet_User() - Sent us a non-NUL terminated New Password.\n", xClient->m_sHostname));
					*bStayAlive = false;
				}
			}
			else
			{
				D(DLOG("[%s] Validate_Packet_User() - Sent us a non-NUL terminated Password.\n", xClient->m_sHostname));
				*bStayAlive = false;
			}
		}
		else
		{
			D(DLOG("[%s] Validate_Packet_User() - Sent us a non-NUL terminated Username.\n", xClient->m_sHostname));
			*bStayAlive = false;
		}
	}
	D(else DLOG("[%s] Validate_Packet_User() - Datalength isn't sufficient for a user.\n", xClient->m_sHostname));
	return bRet;
}

bool Validate_Packet_Group(struct Client *xClient, unsigned int *nAmountRead, bool *bStayAlive)
{
	bool bRet = false;
	if ((*nAmountRead) >= sizeof(struct faPacket_Group))
	{
		unsigned int i;
		xClient->m_bGotCommand = false;
		memcpy(&xClient->m_xPkt_Group, &xClient->m_sBuffer, sizeof(struct faPacket_Group));
		*nAmountRead -= sizeof(struct faPacket_Group);
		memmove(xClient->m_sBuffer, &xClient->m_sBuffer[sizeof(struct faPacket_Group)], *nAmountRead);
		// Validate the packet contents.
		i = 0;
		while ((i < (sizeof(xClient->m_xPkt_Group.m_sGroupname)-1)) && (xClient->m_xPkt_Group.m_sGroupname[i] !='\0')) i++;
		if (i < (sizeof(xClient->m_xPkt_Group.m_sGroupname)-1))
		{
			i = 0;
			while ((i < (sizeof(xClient->m_xPkt_Group.m_sPassword)-1)) && (xClient->m_xPkt_Group.m_sPassword[i] !='\0')) i++;
			if (i < (sizeof(xClient->m_xPkt_Group.m_sPassword)-1))
			{
				// Fix endianness.
				xClient->m_xPkt_Group.m_nGID = ntohl(xClient->m_xPkt_Group.m_nGID);
				xClient->m_xPkt_Group.m_nNumMembers = ntohl(xClient->m_xPkt_Group.m_nNumMembers);
				bRet = true;
			}
			else
			{
				D(DLOG("[%s] Validate_Packet_Group() - Sent us a non-NUL terminated Password.\n", xClient->m_sHostname));
				*bStayAlive = false;
			}
		}
		else
		{
			D(DLOG("[%s] Validate_Packet_Group() - Sent us a non-NUL terminated Groupname.\n", xClient->m_sHostname));
			*bStayAlive = false;
		}
	}
	D(else DLOG("[%s] Validate_Packet_Group() - Datalength isn't sufficient for a user.\n", xClient->m_sHostname));
	return bRet;
}

/*****************************************************************************
 Handle Client
******************************************************************************
 Note: Commands are handled per recv(), meaning that if 200 commands arive
 in one recv() call they all will be handled before going to the next client.
 Keeping the BUFFER_SIZE small prevents this from hogging the other clients
 if one is being abused.
******************************************************************************/
bool HandleClient(struct Client *xClient)
{
	bool				bStayAlive = true;
	unsigned int	nAmountRead = 0;

	D(DLOG("[%s] Handleclient()\n", xClient->m_sHostname));

	// The 'else' shouldn't be possible, but it's just a you-never-know safety precaution.
	if (xClient->m_nOffset < sizeof(xClient->m_sBuffer))
	{
		if ((xClient->m_nMode == c_nMode_Human) || (xClient->m_nMode == c_nMode_Packet))
		{
			int nResult = recv(xClient->m_xSocket, xClient->m_sBuffer + xClient->m_nOffset, sizeof(xClient->m_sBuffer) - xClient->m_nOffset - 8 - 1, 0);
			if (nResult == SOCKET_ERROR) fuzzd_Log(LOG_ERR, "[%s] Handleclient() - recv() failed with error %d: %s\n", xClient->m_sHostname, ERRORNUMSTR);
			else if (nResult == 0) fuzzd_Log(LOG_ERR, "[%s] Handleclient() - Client didn't send us any data.\n", xClient->m_sHostname);
			else nAmountRead = (unsigned)nResult;
		}
		else
		{
			// Packet/SSL or Packet/SSH recv()&decode, but that's not implemented (yet ;)
			D(DLOG("[%s] Handleclient() - Receival of Packet Type %d (%d=%s, %d=%s) Not implemented (yet).\n", xClient->m_sHostname, xClient->m_nMode, c_nMode_Packet_SSH, c_sMode_Packet_SSH, c_nMode_Packet_SSL, c_sMode_Packet_SSL));
			nAmountRead = 0;
		}

		if (nAmountRead == 0)
		{
			bStayAlive = false;
		}
		else
		{
			if (xClient->m_nMode == c_nMode_Human)
			{
				unsigned int i;
				char c;
				//D(DLOG("Read %d bytes\n", nAmountRead));
				for (i=0;  (!g_bQuit) && (bStayAlive) && (i<nAmountRead); i++)
				{
					c = xClient->m_sBuffer[xClient->m_nOffset+i];
					//D(DLOG("Checking '%c' at %d/%d\n", (int)c, i, nAmountRead));
					if ((c == '\n') || (c == '\0'))
					{
						//D(DLOG("Found a string end\n"));
						// We got ourselves a string, lets terminate it.
						if (((xClient->m_nOffset+1)>0)&&(xClient->m_sBuffer[xClient->m_nOffset+i-1]=='\r')) xClient->m_sBuffer[xClient->m_nOffset+i-1]='\0';
						else xClient->m_sBuffer[xClient->m_nOffset+i]='\0';
						bStayAlive = Client_Plaintext_Command(xClient);

						// Move the rest of the buffer to the front.
						memmove(xClient->m_sBuffer, &xClient->m_sBuffer[xClient->m_nOffset+i+1], nAmountRead-i-1);
						// Reset the buffer offset.
						xClient->m_nOffset = 0;
						nAmountRead -= (i+1);
						if (nAmountRead == 0) xClient->m_sBuffer[0]='\0';
						/// Try the buffer again for more commands.
						i=0;
					}
				}
				// We did parse this already but we still need it in our string as it isn't a complete line.
				xClient->m_nOffset = nAmountRead;
			}
			else if ((xClient->m_nMode == c_nMode_Packet) ||
						(xClient->m_nMode == c_nMode_Packet_SSH) ||
						(xClient->m_nMode == c_nMode_Packet_SSL))
			{
				bool bTryMore = true;
				D(char *sName = c_sCommand_Unknown);

				nAmountRead += xClient->m_nOffset;
				xClient->m_nOffset = 0;
				while ((!g_bQuit) && (bStayAlive) && (bTryMore))
				{
					// Do we have a valid Request Packet ?
					if (xClient->m_bGotCommand == false)
					{
						if ((nAmountRead) >= sizeof(struct faPacket_Request))
						{
							// Store this packet in our structure for easy access (and so we keep it even if we cycle the buffer)		
							memcpy(&xClient->m_xPkt_Request, &xClient->m_sBuffer, sizeof(struct faPacket_Request));

							// We have taken this from the buffer, so move the rest of the buffer to the front.
							nAmountRead -= sizeof(struct faPacket_Request);
							memmove(xClient->m_sBuffer, &xClient->m_sBuffer[sizeof(struct faPacket_Request)], nAmountRead);

							// Fixup endianness in the Request packet.
							xClient->m_xPkt_Request.m_nVersion = ntohl(xClient->m_xPkt_Request.m_nVersion);
							xClient->m_xPkt_Request.m_nCommand = ntohl(xClient->m_xPkt_Request.m_nCommand);
							xClient->m_xPkt_Request.m_nDatatype = ntohl(xClient->m_xPkt_Request.m_nDatatype);
							xClient->m_xPkt_Request.m_nLength = ntohl(xClient->m_xPkt_Request.m_nLength);

							// Check the validity of the packet
							// Ofcourse anyone/thing could screw this up for us if they'd/it really wanted.
							// This does take care of nmap and other tools sending unexpected crap to us.
							if	((xClient->m_xPkt_Request.m_nMagic[0]	== 'F') && 
								 (xClient->m_xPkt_Request.m_nMagic[1]	== 'U') &&
								 (xClient->m_xPkt_Request.m_nMagic[2]	== 'Z') &&
								 (xClient->m_xPkt_Request.m_nMagic[3]	== 'Z') &&
								 (xClient->m_xPkt_Request.m_nVersion	== FUZZAUTH_VERSION))
							{
D(
								switch (xClient->m_xPkt_Request.m_nCommand)
								{
									case c_nCommand_Mode:				sName = c_sCommand_Mode;					break;
									case c_nCommand_Quit:				sName = c_sCommand_Quit;					break;
									case c_nCommand_Authenticate:		sName = c_sCommand_Authenticate;			break;
									case c_nCommand_ChangePassword:	sName = c_sCommand_ChangePassword;		break;
									case c_nCommand_GetUsers:			sName = c_sCommand_GetUsers;				break;
									case c_nCommand_MatchUserID:		sName = c_sCommand_MatchUserID;			break;
									case c_nCommand_MatchUserGECOS:	sName = c_sCommand_MatchUserGECOS;		break;
									case c_nCommand_MatchUserName:	sName = c_sCommand_MatchUserName;		break;
									case c_nCommand_GetGroups:			sName = c_sCommand_GetGroups;				break;
									case c_nCommand_MatchGroupID:		sName = c_sCommand_MatchGroupID;			break;
									case c_nCommand_MatchGroupName:	sName = c_sCommand_MatchGroupName;		break;
									case c_nCommand_UserAdd:			sName = c_sCommand_UserAdd;				break;
									case c_nCommand_UserRemove:		sName = c_sCommand_UserRemove;			break;
									case c_nCommand_UserChange:		sName = c_sCommand_UserChange;			break;
									case c_nCommand_GroupAdd:			sName = c_sCommand_GroupAdd;				break;
									case c_nCommand_GroupRemove:		sName = c_sCommand_GroupRemove;			break;
									case c_nCommand_GroupChange:		sName = c_sCommand_GroupChange;			break;
									case c_nCommand_MemberAdd:			sName = c_sCommand_MemberAdd;				break;
									case c_nCommand_MemberRemove:		sName = c_sCommand_MemberRemove;			break;
									case c_nCommand_Unknown:
																				sName = c_sCommand_Unknown;				break;
								}
								DLOG("[%s] RequestPacket() - RequestPacket received; Version: %d, Command: %d -> %s\n", xClient->m_sHostname, xClient->m_xPkt_Request.m_nVersion, xClient->m_xPkt_Request.m_nCommand, sName);
)
								/* Fill in the answer block. */
								xClient->m_xPkt_Answer.m_nAnswer = c_nAnswer_UnknownCommand;
								xClient->m_xPkt_Answer.m_nLength = 0;
								xClient->m_xPkt_Answer.m_nDatatype = c_nDatatype_Unknown;

								/* Handle command */
								switch (xClient->m_xPkt_Request.m_nCommand)
								{
									case c_nCommand_Quit:				xClient->m_xPkt_Answer.m_nAnswer = c_nAnswer_Granted;
																				bStayAlive = false;
																				break;
									case c_nCommand_Mode:
									case c_nCommand_Authenticate:
									case c_nCommand_ChangePassword:
									case c_nCommand_MatchUserID:
									case c_nCommand_MatchUserName:
									case c_nCommand_MatchUserGECOS:
									case c_nCommand_MatchGroupID:
									case c_nCommand_MatchGroupName:
									case c_nCommand_UserAdd:
									case c_nCommand_UserRemove:
									case c_nCommand_UserChange:
									case c_nCommand_GroupAdd:
									case c_nCommand_GroupRemove:
									case c_nCommand_GroupChange:
									case c_nCommand_MemberAdd:
									case c_nCommand_MemberRemove:
																				xClient->m_xPkt_Answer.m_nAnswer = c_nAnswer_NeedDataPacket;
																				break;
									case c_nCommand_GetUsers:			bStayAlive = Cmd_GetUsers(xClient);
																				D(if (!bStayAlive) DLOG("[%s] Handleclient() - Cmd_GetUsers() wants me to end this client\n", xClient->m_sHostname));
																				break;
									case c_nCommand_GetGroups:			bStayAlive = Cmd_GetGroups(xClient);
																				D(if (!bStayAlive) DLOG("[%s] Handleclient() - Cmd_GetGroups() wants me to end this client\n", xClient->m_sHostname));
																				break;
									case c_nCommand_Unknown:
#ifndef DEBUG
									default:
#endif
																				fuzzd_Log(LOG_ERR, "[%s] Handleclient() - Unknown command %d\n", xClient->m_sHostname, xClient->m_xPkt_Request.m_nCommand);
																				xClient->m_xPkt_Answer.m_nAnswer = c_nAnswer_UnknownCommand;
																				break;
								}
D(
								if (xClient->m_xPkt_Request.m_nCommand > c_nCommand_MemberRemove)
								{
									xClient->m_xPkt_Answer.m_nAnswer = c_nAnswer_UnknownCommand;
								}
)
								if (xClient->m_xPkt_Answer.m_nAnswer != c_nAnswer_NeedDataPacket)
								{
									if (xClient->m_xPkt_Request.m_nCommand != c_nCommand_Quit)
									{
										bStayAlive = fuzzd_ReturnAnswer(xClient);
										D(if (!bStayAlive) DLOG("[%s] Handleclient() - Cmd_GetUsers() wants me to end this client\n", xClient->m_sHostname));
									}
									else fuzzd_ReturnAnswer(xClient);
								}
								else
								{
									xClient->m_bGotCommand = true;
									D(DLOG("[%s] Handleclient() - We got a command, now all we want is a data packet...\n", xClient->m_sHostname));
								}
							}
							else
							{
								if (xClient->m_xPkt_Request.m_nVersion	!= FUZZAUTH_VERSION) fuzzd_Log(LOG_ERR, "[%s] Handleclient() - Wrong protocol version number (got: %d, want: %d) detected... killing client.\n", xClient->m_sHostname, xClient->m_xPkt_Request.m_nVersion, FUZZAUTH_VERSION);
								else fuzzd_Log(LOG_ERR, "[%s] Handleclient() - Received wrong protocol magic header...\n", xClient->m_sHostname);
								bStayAlive = false;
							}
						}
						else
						{
							bTryMore = false;
							D(DLOG("[%s] Handleclient() - We are out of data...#1\n", xClient->m_sHostname));
						}
					}
					else // We've got a command now get ourselves the data for it.
					{
						D(DLOG("[%s] Handleclient() - We got a command and probably also a datapacket of size %d\n", xClient->m_sHostname, nAmountRead));
						switch (xClient->m_xPkt_Request.m_nCommand)
						{
							case c_nCommand_Mode:
																// Validate & Copy the packet from our buffer.
																if (Validate_Packet_Number(xClient, &nAmountRead))
																{
																	// Currently we deny all mode changes.
																	// As we haven't implemented any others ;)
																	xClient->m_xPkt_Answer.m_nAnswer = c_nAnswer_Denied;
																}
																else bTryMore = false;
																break;
							case c_nCommand_Authenticate:
																// Validate & Copy the packet from our buffer.
																if (Validate_Packet_User(xClient, &nAmountRead, &bStayAlive))
																{
																	// Authenticate()
																	if (Cmd_Authenticate(xClient)) xClient->m_xPkt_Answer.m_nAnswer = c_nAnswer_Granted;
																	else xClient->m_xPkt_Answer.m_nAnswer = c_nAnswer_Denied;
																}
																else bTryMore = false;
																break;
							case c_nCommand_ChangePassword:
																// Validate & Copy the packet from our buffer.
																if (Validate_Packet_User(xClient, &nAmountRead, &bStayAlive))
																{
																	// Authenticate()
																	if (Cmd_ChangePassword(xClient)) xClient->m_xPkt_Answer.m_nAnswer = c_nAnswer_Granted;
																	else xClient->m_xPkt_Answer.m_nAnswer = c_nAnswer_Denied;
																}
																else bTryMore = false;
																break;
							case c_nCommand_MatchUserID:
																// Validate & Copy the packet from our buffer.
																if (Validate_Packet_Number(xClient, &nAmountRead))
																{
																	// Match, will return a User.
																	bStayAlive = Cmd_MatchUserID(xClient);
																	D(if (!bStayAlive) DLOG("[%s] Handleclient() - Cmd_MatchUserID() wants me to end this client\n", xClient->m_sHostname));
																}
																else bTryMore = false;
																break;
							case c_nCommand_MatchUserName:
																// Validate & Copy the packet from our buffer.
																if (Validate_Packet_String(xClient, &nAmountRead, &bStayAlive))
																{
																	// Match, will return a User.
																	bStayAlive = Cmd_MatchUserName(xClient);
																	D(if (!bStayAlive) DLOG("[%s] Handleclient() - Cmd_MatchUserName() wants me to end this client\n", xClient->m_sHostname));
																}
																else bTryMore = false;
																break;
							case c_nCommand_MatchUserGECOS:
																// Validate & Copy the packet from our buffer.
																if (Validate_Packet_String(xClient, &nAmountRead, &bStayAlive))
																{
																	// Match, will return a User.
																	bStayAlive = Cmd_MatchUserGECOS(xClient);
																	D(if (!bStayAlive) DLOG("[%s] Handleclient() - Cmd_MatchUserGECOS() wants me to end this client\n", xClient->m_sHostname));
																}
																else bTryMore = false;
																break;
							case c_nCommand_MatchGroupID:
																// Validate & Copy the packet from our buffer.
																if (Validate_Packet_Number(xClient, &nAmountRead))
																{
																	// Match, will return a Group packet.
																	bStayAlive = Cmd_MatchGroupID(xClient);
																	D(if (!bStayAlive) DLOG("[%s] Handleclient() - Cmd_MatchGroupID() wants me to end this client\n", xClient->m_sHostname));
																}
																else bTryMore = false;
																break;
							case c_nCommand_MatchGroupName:
																// Validate & Copy the packet from our buffer.
																if (Validate_Packet_String(xClient, &nAmountRead, &bStayAlive))
																{
																	// Match, will return a Group packet.
																	bStayAlive = Cmd_MatchGroupName(xClient);
																	D(if (!bStayAlive) DLOG("[%s] Handleclient() - Cmd_MatchGroupName() wants me to end this client\n", xClient->m_sHostname));
																}
																else bTryMore = false;
																break;
							case c_nCommand_UserAdd:
																// Validate & Copy the packet from our buffer.
																if (Validate_Packet_User(xClient, &nAmountRead, &bStayAlive))
																{
																	// Match, will return a Group packet.
																	bStayAlive = Cmd_UserAdd(xClient);
																	D(if (!bStayAlive) DLOG("[%s] Handleclient() - Cmd_UserAdd() wants me to end this client\n", xClient->m_sHostname));
																}
																else bTryMore = false;
																break;
							case c_nCommand_UserRemove:
																// Validate & Copy the packet from our buffer.
																if (Validate_Packet_String(xClient, &nAmountRead, &bStayAlive))
																{
																	// Match, will return a Group packet.
																	bStayAlive = Cmd_UserRemove(xClient);
																	D(if (!bStayAlive) DLOG("[%s] Handleclient() - Cmd_UserRemove() wants me to end this client\n", xClient->m_sHostname));
																}
																else bTryMore = false;
																break;
							case c_nCommand_UserChange:
																// Validate & Copy the packet from our buffer.
																if (Validate_Packet_User(xClient, &nAmountRead, &bStayAlive))
																{
																	// Match, will return a Group packet.
																	bStayAlive = Cmd_UserChange(xClient);
																	D(if (!bStayAlive) DLOG("[%s] Handleclient() - Cmd_UserChange() wants me to end this client\n", xClient->m_sHostname));
																}
																else bTryMore = false;
																break;
							case c_nCommand_GroupAdd:
																// Validate & Copy the packet from our buffer.
																if (Validate_Packet_Group(xClient, &nAmountRead, &bStayAlive))
																{
																	// Match, will return a Group packet.
																	bStayAlive = Cmd_GroupAdd(xClient);
																	D(if (!bStayAlive) DLOG("[%s] Handleclient() - Cmd_GroupAdd() wants me to end this client\n", xClient->m_sHostname));
																}
																else bTryMore = false;
																break;
							case c_nCommand_GroupRemove:
																// Validate & Copy the packet from our buffer.
																if (Validate_Packet_String(xClient, &nAmountRead, &bStayAlive))
																{
																	// Match, will return a Group packet.
																	bStayAlive = Cmd_MemberRemove(xClient);
																	D(if (!bStayAlive) DLOG("[%s] Handleclient() - Cmd_MemberRemove() wants me to end this client\n", xClient->m_sHostname));
																}
																else bTryMore = false;
																break;
							case c_nCommand_GroupChange:
																// Validate & Copy the packet from our buffer.
																if (Validate_Packet_Group(xClient, &nAmountRead, &bStayAlive))
																{
																	// Match, will return a Group packet.
																	bStayAlive = Cmd_GroupChange(xClient);
																	D(if (!bStayAlive) DLOG("[%s] Handleclient() - Cmd_GroupChange() wants me to end this client\n", xClient->m_sHostname));
																}
																else bTryMore = false;
																break;
							case c_nCommand_MemberAdd:
																// Validate & Copy the packet from our buffer.
																if (Validate_Packet_Number2(xClient, &nAmountRead))
																{
																	// Match, will return a Group packet.
																	bStayAlive = Cmd_MemberAdd(xClient);
																	D(if (!bStayAlive) DLOG("[%s] Handleclient() - Cmd_MemberAdd() wants me to end this client\n", xClient->m_sHostname));
																}
																else bTryMore = false;
																break;
							case c_nCommand_MemberRemove:
																// Validate & Copy the packet from our buffer.
																if (Validate_Packet_Number2(xClient, &nAmountRead))
																{
																	// Match, will return a Group packet.
																	bStayAlive = Cmd_MemberRemove(xClient);
																	D(if (!bStayAlive) DLOG("[%s] Handleclient() - Cmd_MemberRemove() wants me to end this client\n", xClient->m_sHostname));
																}
																else bTryMore = false;
																break;

																// The cases which aren't handled here... this is to make gcc happy when we use the case-enum check.
							case c_nCommand_Unknown:
							case c_nCommand_Quit:
							case c_nCommand_GetUsers:
							case c_nCommand_GetGroups:
D(
							default:
)
																break;
					}
D(
						if (xClient->m_xPkt_Request.m_nCommand > c_nCommand_MemberRemove)
						{
							xClient->m_xPkt_Answer.m_nAnswer = c_nAnswer_UnknownCommand;
							bStayAlive = false;
							bTryMore = false;
							xClient->m_bGotCommand = false;
						}
)
						if (xClient->m_xPkt_Answer.m_nAnswer != c_nAnswer_NeedDataPacket)
						{
							bStayAlive = fuzzd_ReturnAnswer(xClient);
							D(if (!bStayAlive) DLOG("[%s] Handleclient() - ReturnAnswer() wants me to end this client\n", xClient->m_sHostname));
						}
						else
						{
							xClient->m_bGotCommand = true;
							D(DLOG("[%s] Handleclient() - We got a command, now all we want is a data packet...\n", xClient->m_sHostname));
						}
					}
					if (nAmountRead == 0)
					{
						D(DLOG("[%s] Handleclient() - We are out of data...#2\n", xClient->m_sHostname));
						bTryMore = false;
					}
				} // while (!g_bQuit) && (bTryMore)
				// We did parse this already but we still need some more for a complete packet.
				xClient->m_nOffset=nAmountRead;
			}
			else
			{
				fuzzd_Log(LOG_ERR, "[%s] Handleclient() - Unsupported protocol %d specified.\n", xClient->m_sHostname, xClient->m_nMode);
				bStayAlive = false;
			}
		} // if ((nAmountRead == SOCKET_ERROR) || (nAmountRead == 0))
	} //if (xClient->m_nOffset < sizeof(xClient->m_sBuffer))
	else
	{
		fuzzd_Log(LOG_ERR, "[%s] Handleclient() - Killing client which filled the buffer\n", xClient->m_sHostname);
		bStayAlive = false;
	}
	return bStayAlive;
}

// Set us up with a UNIX socket,
// this is for the communication by local clients.
bool Setup_UnixSocket(bool bRetry)
{
	bool bRet = false;
	struct stat xStat;

	if (lstat(FUZZAUTH_VARDIR, &xStat) == 0)
	{
		// Is it a directory?
		if (S_ISDIR(xStat.st_mode))
		{
			// The owner of the directory should be 'root' (uid=0)
			// It should have the same group as us.
			if ((xStat.st_uid == 0) && (xStat.st_gid == getgid()))
			{
				// And we and root should be able to Read,Write,Explore to it, others should be able to Read+eXplore it.
				if ((xStat.st_mode & 0777) == 0775)
				{
					g_xServSock[0] = socket(PF_UNIX, SOCK_STREAM, 0);
					if (g_xServSock[0] != INVALID_SOCKET)
					{
						struct sockaddr_un xSockAddrUnix;

						/* Make sure everybody can write to this socket. */
						mode_t m = umask(0x000);

						memset(&xSockAddrUnix, 0, sizeof(xSockAddrUnix));
						xSockAddrUnix.sun_family = AF_UNIX;
						strncpy(xSockAddrUnix.sun_path, FUZZAUTH_UNIXSOCKET, sizeof(xSockAddrUnix.sun_path) - 1);

						if (bind(g_xServSock[0], (struct sockaddr *)&xSockAddrUnix, sizeof(xSockAddrUnix)) != SOCKET_ERROR)
						{
							if (listen(g_xServSock[0], 5) != SOCKET_ERROR)
							{
								bRet = true;
								fuzzd_Log(LOG_INFO, "Listening on UNIX socket called %s\n", FUZZAUTH_UNIXSOCKET);
							}
							else
							{
								fuzzd_Log(LOG_ERR, "listen() failed on UNIX socket %s: %d: %s\n", FUZZAUTH_UNIXSOCKET, ERRORNUMSTR);
								closesocket(g_xServSock[0]);
								g_xServSock[0] = INVALID_SOCKET;
							}
						}
						else
						{
							if (bRetry)
							{
								if (lstat(FUZZAUTH_UNIXSOCKET, &xStat) == 0)
								{
									if (S_ISSOCK(xStat.st_mode))
									{
										if ((xStat.st_uid == getuid()) && (xStat.st_gid == getgid()))
										{
											SOCKET sSocket = socket(PF_UNIX, SOCK_STREAM, 0);
											fuzzd_Log(LOG_NOTICE, "Checking if there is another FUZZd running\n");
											if (sSocket != INVALID_SOCKET)
											{
												struct sockaddr_un xSockAddrUnix;
												memset(&xSockAddrUnix, 0, sizeof(xSockAddrUnix));
												xSockAddrUnix.sun_family = AF_UNIX;
												strncpy(xSockAddrUnix.sun_path, FUZZAUTH_UNIXSOCKET, sizeof(xSockAddrUnix.sun_path) - 1);
					
												if (connect(sSocket, (struct sockaddr *)&xSockAddrUnix, sizeof(xSockAddrUnix)) != SOCKET_ERROR)
												{
													// Apparently this FUZZd is still alive.
													fuzzd_Log(LOG_NOTICE, "Old working FUZZd detected... aborting startup of new one.\n");
												}
												else
												{
													// Apparently this is a killed/failed/crashed/whatevered one.
													fuzzd_Log(LOG_NOTICE, "Old FUZZd was not alive... killing old socket, creating my own.\n");
					
													// Delete the lingering socket.
													unlink(FUZZAUTH_UNIXSOCKET);
					
													// And try again... but don't do a retry this time :)
													Setup_UnixSocket(false);
												}
												closesocket(sSocket);
												sSocket = INVALID_SOCKET;
											}
											else fuzzd_Log(LOG_ERR, "Couldn't get an AF_UNIX socket\n");
										}
										else fuzzd_Log(LOG_ERR, "The socket (\"%s\") is owned by someone else (%d:%d) than me (%d:%d), I don't like that.\n", FUZZAUTH_UNIXSOCKET, xStat.st_uid, xStat.st_gid, getuid(), getgid());
									}
									else fuzzd_Log(LOG_ERR, "Old \"socket\" apparently isn't a socket, strange things going on here, giving up.\n");
								}
								else fuzzd_Log(LOG_ERR, "Couldn't stat old socket, strange things going on here, giving up.\n");
							}
							else
							{
								fuzzd_Log(LOG_ERR, "bind() failed on UNIX socket %s: %d: %s\n", FUZZAUTH_UNIXSOCKET, ERRORNUMSTR);
								closesocket(g_xServSock[0]);
								g_xServSock[0] = INVALID_SOCKET;
							}
						}
						umask(m);
					}
					else fuzzd_Log(LOG_ERR, "Couldn't create UNIX socket %s: %d: %s\n", FUZZAUTH_UNIXSOCKET, ERRORNUMSTR);
				}
				else fuzzd_Log(LOG_ERR, "The permissions for the variable dir (\"%s\") are wrong, currently %o, but they should be %o.\n", FUZZAUTH_VARDIR, xStat.st_mode & 0777, 0775);
			}
			else fuzzd_Log(LOG_ERR, "The variable dir (\"%s\") is owned by someone else (%d:%d) than (%d:%d), I don't like that.\n", FUZZAUTH_VARDIR, xStat.st_uid, xStat.st_gid, 0, getgid());
		}
		else fuzzd_Log(LOG_ERR, "The variable dir (\"%s\") isn't a directory.\n", FUZZAUTH_VARDIR);
	}
	else fuzzd_Log(LOG_ERR, "The variable dir (\"%s\"), where the socket needs to go doesn't exist.\n", FUZZAUTH_VARDIR);
	return bRet;
}

void Cleanup_UnixSocket(void)
{
	closesocket(g_xServSock[0]);
	// And remove the file just in case.
	unlink(FUZZAUTH_UNIXSOCKET);
}

void Setup_NetSockets(int nFamily, char *sAddress, char *sPort)
{
	bool					bListeningOnIPv6 = false;
	struct addrinfo	xHints, *xAddrInfo, *xAI;
	int					nRetVal;

	FD_SET(g_xServSock[0], &g_xSockSet);
	net_set_non_blocking(g_xServSock[0]);

	// By setting the AI_PASSIVE flag in the hints to getaddrinfo, we're
	// indicating that we intend to use the resulting address(es) to bind
	// to a socket(s) for accepting incoming connections.  This means that
	// when the Address parameter is NULL, getaddrinfo will return one
	// entry per allowed protocol family containing the unspecified address
	// for that family.
	D(DLOG("Trying to getaddrinfo() address %s, port %s, protocol family %s\n", sAddress, sPort, (xHints.ai_family == PF_INET) ? "PF_INET" : "PF_INET6"));
	memset(&xHints, 0, sizeof(xHints));
	xHints.ai_family = nFamily;
	xHints.ai_socktype = SOCK_STREAM;
	xHints.ai_flags = AI_NUMERICHOST | AI_PASSIVE;
	nRetVal = getaddrinfo(sAddress, sPort, &xHints, &xAddrInfo);
	if (nRetVal == 0)
	{
		// For each address getaddrinfo returned, we create a new socket,
		// bind that address to it, and create a queue to listen on.
		for (xAI = xAddrInfo; (g_nNumSocks < FD_SETSIZE) && (xAI != NULL); xAI = xAI->ai_next)
		{
			// We only support PF_INET and PF_INET6.
			if ((xAI->ai_family == PF_INET) || (xAI->ai_family == PF_INET6))
			{
				// Open a socket with the correct address family for this address.
				g_xServSock[g_nNumSocks] = socket(xAI->ai_family, xAI->ai_socktype, xAI->ai_protocol);
				if (g_xServSock[g_nNumSocks] != INVALID_SOCKET)
				{
					//
					// bind() associates a local address and port combination
					// with the socket just created. This is most useful when
					// the application is a server that has a well-known port
					// that clients know about in advance.
					//
					if (bind(g_xServSock[g_nNumSocks], xAI->ai_addr, xAI->ai_addrlen) != SOCKET_ERROR)
					{
						if (listen(g_xServSock[g_nNumSocks], 5) != SOCKET_ERROR)
						{
							fuzzd_Log(LOG_INFO, "Listening on port %s, protocol family %s\n", sPort, (xAI->ai_family == PF_INET) ? "PF_INET" : "PF_INET6");
							if (g_xServSock[g_nNumSocks] > g_nHighSock) g_nHighSock = g_xServSock[g_nNumSocks];
							FD_SET(g_xServSock[g_nNumSocks], &g_xSockSet);
							net_set_non_blocking(g_xServSock[g_nNumSocks]);
							g_nNumSocks++;
							if (xAI->ai_family == PF_INET6) bListeningOnIPv6 = true;
						}
						else
						{
							fuzzd_Log(LOG_ERR, "listen() failed with %d: %s\n", ERRORNUMSTR);
							closesocket(g_xServSock[g_nNumSocks]);
							g_xServSock[g_nNumSocks] = INVALID_SOCKET;
							errno = 0;
						}
					}
					else
					{
						// If we are already listening on IPv6, on some implementations this also implies IPv4 so don't log this case with as an error.
						fuzzd_Log((xAI->ai_family == PF_INET) && (bListeningOnIPv6 == true) ? LOG_INFO : LOG_ERR, "bind() failed for protocol family %s with error %d: %s%s\n", (xAI->ai_family == PF_INET) ? "PF_INET" : "PF_INET6", ERRORNUMSTR, (xAI->ai_family == PF_INET) && (bListeningOnIPv6 == true) ? " (But we are already listening on IPv6 so that could be it ;)" : "");
						closesocket(g_xServSock[g_nNumSocks]);
						g_xServSock[g_nNumSocks] = INVALID_SOCKET;
						errno = 0;
					}
				}
				else fuzzd_Log(LOG_ERR, "socket() failed with error %d: %s\n", ERRORNUMSTR);
			}
			D(else DLOG("Unknown address family %d\n", xAI->ai_family));
		}
		if (g_nNumSocks >= FD_SETSIZE) fuzzd_Log(LOG_ERR, "getaddrinfo returned more addresses than we could use, ignoring the rest.\n");
		freeaddrinfo(xAddrInfo);
	}
	else fuzzd_Log(LOG_ERR, "getaddrinfo failed with error %d: %s\n", nRetVal, gai_strerror(nRetVal));
}

void Cleanup_NetSockets(void)
{
	int i;
	/* Cleanup g_xServSock[1+] */
	for (i = 1; i < g_nNumSocks; i++)
	{
		closesocket(g_xServSock[i]);
		g_xServSock[i] = INVALID_SOCKET;
	}
}

void MainLoop(void)
{
	struct timeval				xSelectMaxTime;
	fd_set						xSockSetReady;
	struct node					*xNode, *xNode2;
	struct Client				*xClient;
	int							i;

	FD_ZERO(&xSockSetReady);

	while (!g_bQuit)
	{
		memcpy(&xSockSetReady, &g_xSockSet, sizeof(xSockSetReady));
		xSelectMaxTime.tv_sec	= 10000;
		xSelectMaxTime.tv_usec	= 0;
		i = select(g_nHighSock + 1, &xSockSetReady, 0, 0, &xSelectMaxTime);
		if ((!g_bQuit) && (i != -1))
		{
			for (i = 0; i < g_nNumSocks; i++)
			{
				if (FD_ISSET(g_xServSock[i], &xSockSetReady))
				{
					FD_CLR(g_xServSock[i], &xSockSetReady);
					AcceptClient(g_xServSock[i], i==0?true:false);
				}
			}
			xNode = lst_GetFirst(&g_xClientList);
			while ((!g_bQuit) && (xNode))
			{
				xNode2 = xNode;
				xNode = lst_GetNext(xNode);
				xClient = (struct Client *)lst_GetData(xNode2);
				if (xClient)
				{
					if (FD_ISSET(xClient->m_xSocket, &xSockSetReady))
					{
						FD_CLR(xClient->m_xSocket, &xSockSetReady);
						if (!HandleClient(xClient))
						{
							if (xClient->m_xSocket == g_nHighSock) g_nHighSock--;
							FD_CLR(xClient->m_xSocket, &g_xSockSet);
							closesocket(xClient->m_xSocket);
							xClient->m_xSocket = INVALID_SOCKET;
							if ((!xClient->m_bUnix) || ((xClient->m_bUnix) && (g_bVerbose)) D(|| (g_nDebugLevel > 0) ) ) fuzzd_Log(LOG_INFO, "[%s] Client is leaving the arena...\n", xClient->m_sHostname);
							D(DLOG("[%s] =======================================================\n", xClient->m_sHostname));
							free(xClient);
							lst_RemNode(xNode2);
						}
					}
				}
			}
		}
		else fuzzd_Log(LOG_INFO, "select() was interrupted...\n");
	}

	// Cleanup time(tm)
	xNode = lst_GetFirst(&g_xClientList);
	while ((g_bQuit) && (xNode))
	{
		xNode2 = xNode;
		xNode = lst_GetNext(xNode);
		xClient = (struct Client *)lst_GetData(xNode2);
		if (xClient)
		{
			fuzzd_Log(LOG_INFO, "Telling Client %s that the arena is closing...\n", xClient->m_sHostname);
			if (xClient->m_nMode == c_nMode_Human) soprintf(xClient, "500 %s shutting down.\r\n", FUZZAUTH_NAME);
			else
			{
				// Send a close answer.
			}
			free(xClient);
			lst_RemNode(xNode2);
		}
	}
}

/************************
 Usage
************************/
void Usage(char *sProgName, char *sProblem, char *sProblemArg)
{
	printf("%s %s - %s v%d Server by %s <%s>\n", FUZZD_NAME, FUZZAUTH_DATE, FUZZAUTH_NAME, FUZZAUTH_VERSION, FUZZAUTH_AUTHOR, FUZZAUTH_EMAIL);
	printf("Website: %s\n\n", FUZZAUTH_WEBSITE);
#ifdef DEBUG
	printf("%s [-l] [-f family] [-p port] [-a address] [-s hostid] [-b backend] [-P pidfile]] [-v] [-d debuglevel] \n\n", sProgName);
#else
	printf("%s [-l] [-f family] [-p port] [-a address] [-s hostid] [-b backend] [-P pidfile]] [-v]\n\n", sProgName);
#endif
	printf("  -l\t\tOnly listen on the unix socket, don't bind any inet sockets. (default: bind all)\n");
	printf("  family\tOne of PF_INET, PF_INET6 or PF_UNSPEC.  (default: %s)\n",  (DEFAULT_FAMILY == PF_UNSPEC) ? "PF_UNSPEC" : ((DEFAULT_FAMILY == PF_INET) ? "PF_INET" : "PF_INET6"));
	printf("  port\t\tPort on which to bind.  (default: %s)\n", FUZZAUTH_PORT);
	printf("  address\tIP address on which to bind.  (default: unspecified address)\n");
	printf("  hostid\tIdentifier of this server (default: %s)\n", g_sServername);
	printf("  backend\tBackend Authenticator Module (default: %s)\n", FUZZD_AUTH_MODULE);
	printf("  pidfile\tFile to store the process id in of the daemon (default: %s)\n", FUZZAUTH_FUZZD_PID);
	printf("  -v\t\tBe verbose, currently this enables logging of UNIX Socket connections. (default: no verbosity)\n");
D(
	printf("  debuglevel\tThe debuglevel, 0=none, 10=all, also prohibits backgrounding of fuzzd (default: 0).\n");
)

	if (sProblem)
	{
		printf("\nERROR: ");
		printf(sProblem, sProblemArg);
		printf("\n");
	}
	g_bQuit = true;
}

void ParseArguments(int nArgc, char **psArgv, bool *bLocalOnly, int *nFamily, char **sAddress, char **sPort, char **sBackend, char **sPidFile)
{
	int i;

	// Apply some defaults...
	if (strlen(g_sServername) == 0)
	{
		if (gethostname(g_sServername, sizeof(g_sServername) - 1) != 0) strncat(g_sServername, FUZZAUTH_NAME_LOW, sizeof(g_sServername) - 1);
		strncat(g_sServername, ".", sizeof(g_sServername) - strlen(g_sServername) - 1);
		getdomainname(&g_sServername[strlen(g_sServername)], sizeof(g_sServername) - strlen(g_sServername) - 1);
	}

	// Parse arguments
	if (nArgc > 1)
	{
		for (i = 1; (!g_bQuit) && (i < nArgc);i++)
		{
			if ((psArgv[i][0] == '-') || ((psArgv[i][0] == '/') && (psArgv[i][1] != 0) && (psArgv[i][2] == 0)))
			{
				switch (psArgv[i][1])
				{
					case 'l':		*bLocalOnly = true;
										break;
					case 'v':		g_bVerbose = true;
										break;
					case 'f':
										if (psArgv[i+1])
										{
											if (!strcasecmp(psArgv[i+1], "PF_INET")) *nFamily = PF_INET;
											else if (!strcasecmp(psArgv[i+1], "PF_INET6")) *nFamily = PF_INET6;
											else if (!strcasecmp(psArgv[i+1], "PF_UNSPEC")) *nFamily = PF_UNSPEC;
											else Usage(psArgv[0], "Unknown address family (\"%s\") specified for -f", psArgv[i+1]);
										}
										else Usage(psArgv[0], "-f requires an address family.", NULL);
										i++;
										break;
					case 'a':
										if ((psArgv[i+1]) && (psArgv[i+1][0] != '-'))
										{
											*sAddress = psArgv[++i];
										}
										else Usage(psArgv[0], "-a requires a valid local hostname, a IPv4 or IPv6 address.", NULL);
										break;
					case 'p':
										if ((psArgv[i+1]) && (psArgv[i+1][0] != '-'))
										{
											*sPort = psArgv[++i];
										}
										else Usage(psArgv[0], "-p requires a valid portname or number.", NULL);
										break;
D(
					case 'd':
										if ((psArgv[i+1]) && (psArgv[i+1][0] != '-'))
										{
											g_nDebugLevel = atoi(psArgv[i+1]);
											if (g_nDebugLevel <= 0 || g_nDebugLevel > 10) Usage(psArgv[0], "%s is not a valid argument for -d", psArgv[i+1]);
											i++;
										}
										else Usage(psArgv[0], "-d requires a numberic argument ranging from 1-10 specifying the debuglevel.", NULL);
										break;
)
					case 's':
										if (psArgv[i+1])
										{
											strncpy(g_sServername, psArgv[i+1], sizeof(g_sServername));
										}
										else Usage(psArgv[0], "-s requires a name.", NULL);
										i++;
										break;
					case 'b':
										if (psArgv[i+1])
										{
											*sBackend = psArgv[++i];
										}
										else Usage(psArgv[0], "-b requires a backend name.", NULL);
										break;
					case 'P':
										if (psArgv[i+1])
										{
											*sPidFile = psArgv[++i];
										}
										else Usage(psArgv[0], "-P requires a PidFile name.", NULL);
										break;
					default:
										Usage(psArgv[0], "Unknown option (\"%s\") encountered", psArgv[i]);
										break;
				}
			}
			else Usage(psArgv[0], "Unknown option encountered", NULL);
		}
	}
}

/*********
 Main() ;)
**********/
int main(int nArgc, char **psArgv)
{
	char							*sPort = FUZZAUTH_PORT, *sAddress = NULL, *sBackend = NULL, *sPidFile = FUZZAUTH_FUZZD_PID;
	int							nFamily = DEFAULT_FAMILY, j;
	struct passwd				*xPassword;
	struct group				*xGroup;
	bool							bLocalOnly = false;
#ifdef WIN32
	WSADATA						wsaData;
#endif
	lst_New(&g_xClientList);
	g_sServername[0]='\0';

	umask(0077);

#ifdef WIN32
	// Ask for Winsock version 2.2.
	if ((nRetVal = WSAStartup(MAKEWORD(2, 2), &wsaData)) != 0)
	{
		fuzzd_Log(LOG_ERR, "WSAStartup failed with error %d: %s\n", ERRORNUMSTR);
	}
	else
	{
#endif
	// Determine the UID of the user we should be.
	if ((xPassword = getpwnam(FUZZAUTH_NAME_LOW)))
	{
		// Check if we are running under the correct user id.
		if ((getuid() != xPassword->pw_uid) || (geteuid() != xPassword->pw_uid))
		{
			fuzzd_Log(LOG_ERR, "The normal and effective user id should be %d (\"%s\"), but uid=%d, euid=%d\n", xPassword->pw_uid, FUZZAUTH_NAME_LOW, getuid(), geteuid());
			fuzzd_Log(LOG_ERR, "Aborting startup\n");
			g_bQuit = true;
		}
	}
	else
	{
		// We don't exist
		fuzzd_Log(LOG_ERR, "User \"%s\" doesn't exist at all.... read the doc/INSTALL file again!\n", FUZZAUTH_NAME_LOW);
		g_bQuit = true;
	}

	// Determine the GID of the user we should be.
	if ((xGroup = getgrnam(FUZZAUTH_NAME_LOW)))
	{
		// Check if we are running under the correct user id.
		if ((getgid() != xGroup->gr_gid) || (getegid() != xGroup->gr_gid))
		{
			fuzzd_Log(LOG_ERR, "The normal and effective group id should be %d (\"%s\"), but gid=%d and egid=%d\n", xGroup->gr_gid, FUZZAUTH_NAME_LOW, getgid(), getegid());
			fuzzd_Log(LOG_ERR, "Aborting startup\n");
			g_bQuit = true;
		}
	}
	else
	{
		// We don't exist
		fuzzd_Log(LOG_ERR, "Group \"%s\" doesn't exist at all.... read the doc/INSTALL file again!\n", FUZZAUTH_NAME_LOW);
		g_bQuit = true;
	}

	if (!g_bQuit) ParseArguments(nArgc, psArgv, &bLocalOnly, &nFamily, &sAddress, &sPort, &sBackend, &sPidFile);

	if (sBackend == NULL) Usage(psArgv[0], "No backend defined", NULL);

	if (!g_bQuit)
	{
#ifndef CANNOT_DAEMONIZE
		D(if (g_nDebugLevel == 0)) j = fork();
		D(else j = 0;)
		if (j < 0) fuzzd_Log(LOG_ERR, "Aaaaarggghh, couldn't fork, exiting...\n");
		else if (j == 0)
		{
			D(if (g_nDebugLevel == 0)
			{)
				// Our deamonized thread/process
				D(DLOG("Detaching from tty, becoming a daemon\n"));
				setsid();
				chdir("/");
				/* Don't show any output...*/
				freopen("/dev/null", "r", stdin);
				freopen("/dev/null", "w", stdout);
				freopen("/dev/null", "w", stdout);
			D(})
#endif
			// NOTE: Everything should close and clean (NULLify) nicely because we can do another run.
			while (!g_bQuit)
			{
				FD_ZERO(&g_xSockSet);
				g_nHighSock = 0;
				g_nNumSocks = 0;
				fuzzd_Log(LOG_INFO, "%s %s - %s v%d Server by %s <%s>\n", FUZZD_NAME, FUZZAUTH_DATE, FUZZAUTH_NAME, FUZZAUTH_VERSION, FUZZAUTH_AUTHOR, FUZZAUTH_EMAIL);
				fuzzd_Log(LOG_INFO, "Running on %s\n", g_sServername);

				if (Setup_UnixSocket(true))
				{
					char sBack[1024];
					snprintf(sBack, sizeof(sBack), "%s/lib/fuzzauth_%s.so", FUZZAUTH_HOMEDIR, sBackend);
					g_xBackend = dlopen(sBack, RTLD_GLOBAL|RTLD_NOW);
					if (g_xBackend)
					{
						snprintf(sBack, sizeof(sBack), "%s_Authenticator", sBackend);
						g_xAuthenticator = (struct Authenticator *)dlsym(g_xBackend, sBack);
						if (g_xAuthenticator)
						{
							char sBuf[2048];
							g_xAuthenticator->cmd_GetInfo(sBuf, sizeof(sBuf)-1);
							fuzzd_Log(LOG_INFO, "%s", sBuf);
						}
						else fuzzd_Log(LOG_INFO, "Couldn't resolve Backend Authenticator structure symbol (%s)...\n", sBack);
					}
					else fuzzd_Log(LOG_INFO, "Couldn't open the library (%s)...\n", sBack);
					if ((g_xAuthenticator) && (g_xAuthenticator->cmd_Initialize()))
					{
#ifdef SIGPIPE
						/* Ignore PIPE errors */
						signal(SIGPIPE, SIG_IGN);
#endif
						/* Make sure it cleans up ok... If someone really wants it dead it would die anyway... */
						/* SIGHUP makes fuzzd restart */
						signal_init();

						g_nNumSocks = 1;
						if (!bLocalOnly) Setup_NetSockets(nFamily, sAddress, sPort);
						if ((!bLocalOnly) && (g_nNumSocks == 1)) fuzzd_Log(LOG_WARNING, "Warning: Only listening on UNIX socket.\n");
						D(DLOG("We've got %d server-socket%s\n", g_nNumSocks, g_nNumSocks==1?"":"s"));
	
						if (sPidFile)
						{
							FILE *xPidFile = fopen(sPidFile, "w");
							if (xPidFile)
							{
								D(DLOG("PID %d written to PidFile \"%s\"\n", getpid(), sPidFile));
								fprintf(xPidFile, "%d\n", getpid());
								fclose(xPidFile);
							}
							D(else DLOG("PidFile \"%s\" couldn't be opened.\n", sPidFile));
						}
						D(else DLOG("No PidFile specified.\n"));
	
						// We now put the server into an eternal loop (until quittime ;),
						// serving requests as they arrive.
						MainLoop();
	
						// Cleanup time
	
						// Remove PidFile
						if (sPidFile) unlink(sPidFile);
	
						// Cleanup Network sockets.
						Cleanup_NetSockets();
						g_xAuthenticator->cmd_Shutdown();
						g_xAuthenticator = NULL;
					}
					else
					{
						fuzzd_Log(LOG_ERR, "Couldn't initialize backend.\n");
						g_bQuit = true;
					}
					if (g_xBackend)
					{
						dlclose(g_xBackend);
						g_xBackend = NULL;
					}
					// Cleanup Unix socket.
					Cleanup_UnixSocket();
				}
				else g_bQuit = true;
				
				if (g_bSigHUP)
				{
					fuzzd_Log(LOG_INFO, "Restarting after SigHUP...\n");
					g_bSigHUP = false;
					g_bQuit = false;
				}
			}
			fuzzd_Log(LOG_INFO, "Shutdown\n");
#ifndef CANNOT_DAEMONIZE
		}
		D(else DLOG("Mother fork exiting....\n"));
#endif
	} // if (!g_bQuit)
#ifdef WIN32
	}
	WSACleanup();
#endif
	return 0;
}
