/******************************************************************************
  List/Node Management (lst_*)
*******************************************************************************
	- lst_Debug		- Show debug info about a list.
	- lst_New		- Create/Initialize a list.
	# lst_IsEmpty	- Checks if list is empty.
	- lst_AddHead	- Add a node at the head.
	- @lst_AddTail	- Add a node at the tail.
	- lst_RemNode	- Remove a node from a list.
	# lst_RemHead	- Remove the head of a list (uses lst_RemNode).
	# lst_RemTail	- Remove the tail of a list (uses lst_RemNode).
******************************************************************************/
// We won't want to debug this... really :)
#ifdef DEBUG
#undef DEBUG
#endif
#include "list.h"
#include "stdlib.h"

#ifdef DEBUG
void lst_Debug(struct list *lst,char *func)
{
	struct node *nod;
	if (lst)
	{
		nod=lst->next;
		/* Forward */
		printf("lst_Debug(%s): %lx(%lx,%lx,%lx)",func,lst,lst->next,lst->prev,lst->tail);
		while (nod->next)
		{
			printf(" -> %lx(%lx,%lx)",nod,nod->next,nod->prev);
			nod=nod->next;
		}
		/* Reverse */
		printf("\r\nReverse :");
		nod=lst->tail;
		while (nod->prev)
		{
			printf(" <- %lx(%lx,%lx)",nod,nod->next,nod->prev);
			nod=nod->prev;
		}
		printf("\r\n");
	}
	else printf("*****************\r\n*****************\r\nNO LIST\r\n*****************\r\n*****************\r\n");
}
void lst_DebugNode(struct node *node,char *func)
{
	struct node *nod=node;
	if (node)
	{
		/* Forward */
		printf("lst_DebugNode(%s): %lx",func,node);
		while (nod->next)
		{
			printf(" -> %lx(%lx,%lx)",nod,nod->next,nod->prev);
			nod=nod->next;
		}
		/* Reverse */
		printf("\r\nReverse  : %lx",node);
		while (nod->prev)
		{
			printf(" <- %lx(%lx,%lx)",nod,nod->next,nod->prev);
			nod=nod->prev;
		}
		printf("\r\n");
	}
	else printf("*****************\r\n*****************\r\nNO NODE\r\n*****************\r\n*****************\r\n");
}
#endif

void lst_New(struct list *lst)
{
	lst->next=(struct node *)&lst->prev;
	lst->prev=NULL;
	lst->tail=(struct node *)&lst->next;			/* &lst->next == lst */
}

/* Add node to begin of list. */
bool lst_AddHead(struct list *lst,void *data)
{
	struct node *nod = (struct node *)malloc(sizeof(struct node));
	if (lst&&nod)
	{
		nod->data=data;
		lst->next->prev=nod;
		nod->next=lst->next;
		lst->next=nod;
		nod->prev=(struct node *)lst;
	}
	return(nod?true:false);
}

/* Add node to end of list. */
bool lst_AddTail(struct list *lst, void *data)
{
	struct node *nod = (struct node *)malloc(sizeof(struct node));
	if (lst&&nod)
	{
		nod->data=data;
		lst->tail->next=nod;
		nod->prev=lst->tail;
		lst->tail=nod;
		nod->next=(struct node *)&lst->prev;
	}
	lst_Debug(lst, "lst_AddTail");
	return(nod?true:false);
}

/* Remove Node from a list */
void lst_RemNode(struct node *nod)
{
	if (nod)
	{
		nod->prev->next=nod->next;
		nod->next->prev=nod->prev;
		nod->next=nod->prev=NULL;
	}
	free(nod);
	lst_DebugNode(nod, "lst_RemNode");
}

struct node *lst_GetFirst(struct list *lst)
{
	lst_Debug(lst, "lst_GetFirst");
	return lst->next->next ? lst->next : NULL;
}

struct node *lst_GetNext(struct node *nod)
{
	lst_DebugNode(nod, "lst_GetNext");
	return nod->next->next ? nod->next : NULL;
}

void *lst_GetData(struct node *nod)
{
	return nod->data;
}
