/*	FUZZauth - http://unfix.org/projects/fuzzauth & http://fuzzauth.sourceforge.net
 * php_fuzzauth - FUZZauth PHP interface by Jeroen Massar <jeroen@unfix.org>
 *
 *	Giving access from PHP to FUZZauth ;)
 *	Which allows one to:
 *		- authenticate users
 *		- change passwords
 *		- check if users exist
 *	from PHP...
 *	Welcome in the age of webbased management ;)
*/

#include "php_fuzzauth.h"
#include "fuzzauth.h"
#include "fuzzauth_common.h"

// This is needed for the info functions.
#include <ext/standard/info.h>

function_entry fuzzauth_functions[] = {
	PHP_FE(fuzzauth_authenticate, NULL)
	PHP_FE(fuzzauth_changepassword, NULL)
	PHP_FE(fuzzauth_userexists, NULL)
	{NULL, NULL, NULL}
};

zend_module_entry fuzzauth_module_entry = {
	STANDARD_MODULE_HEADER,
	FUZZAUTH_NAME,
	fuzzauth_functions,
	NULL,
	NULL,
	NULL,
	NULL,
	ZEND_MODULE_INFO_N(fuzzauth),
	0,
	STANDARD_MODULE_PROPERTIES,
};

#ifdef COMPILE_DL_FUZZAUTH
ZEND_GET_MODULE(fuzzauth)
#endif

PHP_MINFO_FUNCTION(fuzzauth)
{
	char sVersion[10];
	
	snprintf(sVersion, sizeof(sVersion)-1, "%d", FUZZAUTH_VERSION);
	php_info_print_table_start();
	php_info_print_table_header(2,	FUZZAUTH_NAME" support", 				"enabled");
	php_info_print_table_row(2,		FUZZAUTH_NAME" date",					FUZZAUTH_DATE);
	php_info_print_table_row(2,		FUZZAUTH_NAME" protocol version",	sVersion);
	php_info_print_table_end();
	return;
	zend_module = zend_module;
}

// fuzzauth_Authenticate(string xUsername, string xPassword)
PHP_FUNCTION(fuzzauth_authenticate)
{
	pval	*xUsername = NULL, *xPassword = NULL;
	bool	bRet;

	if ((ZEND_NUM_ARGS() == 2) && (zend_get_parameters(ht, 2, &xUsername, &xPassword) == SUCCESS))
	{
		convert_to_string(xUsername);
		convert_to_string(xPassword);

		bRet = fuzzauth_Authenticate(xUsername->value.str.val, xPassword->value.str.val);

		if (bRet)
		{
			RETVAL_TRUE
		}
		else
		{
			RETVAL_FALSE
		}
	}
	else WRONG_PARAM_COUNT
	return;
	
	// Unused params/arguments
	this_ptr = this_ptr;
	return_value_used = return_value_used;
}

// fuzzauth_ChangePassword(string xUsername, string xPasswordOld, string xPasswordNew)
PHP_FUNCTION(fuzzauth_changepassword)
{
	pval	*xUsername = NULL, *xPasswordOld = NULL, *xPasswordNew = NULL;
	bool	bRet;

	if ((ZEND_NUM_ARGS() == 3) && (zend_get_parameters(ht, 3, &xUsername, &xPasswordOld, &xPasswordNew) == SUCCESS))
	{
		convert_to_string(xUsername);
		convert_to_string(xPasswordNew);
		convert_to_string(xPasswordOld);

		bRet = fuzzauth_ChangePassword(xUsername->value.str.val, xPasswordOld->value.str.val, xPasswordNew->value.str.val);

		if (bRet)
		{
			RETVAL_TRUE
		}
		else
		{
			RETVAL_FALSE
		}
	}
	else WRONG_PARAM_COUNT
	return;
	
	// Unused params/arguments
	this_ptr = this_ptr;
	return_value_used = return_value_used;
}

// fuzzauth_ChangePassword(string xUsername)
PHP_FUNCTION(fuzzauth_userexists)
{
	pval	*xUsername = NULL;
	bool	bRet;

	if ((ZEND_NUM_ARGS() == 1) && (zend_get_parameters(ht, 1, &xUsername) == SUCCESS))
	{
		convert_to_string(xUsername);

		bRet = fuzzauth_UserExists(xUsername->value.str.val);

		if (bRet)
		{
			RETVAL_TRUE
		}
		else
		{
			RETVAL_FALSE
		}
	}
	else WRONG_PARAM_COUNT
	return;

	// Unused params/arguments
	this_ptr = this_ptr;
	return_value_used = return_value_used;
}
